Package: shinychat
Title: Chat UI Component for 'shiny'
Version: 0.4.0
Authors@R: c(
    person("Joe", "Cheng", , "joe@posit.co", role = "aut"),
    person("Carson", "Sievert", , "carson@posit.co", role = "aut"),
    person("Garrick", "Aden-Buie", , "garrick@adenbuie.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0002-7111-0077")),
    person("Barret", "Schloerke", , "barret@posit.co", role = "aut",
           comment = c(ORCID = "0000-0001-9986-114X")),
    person("Posit Software, PBC", role = c("cph", "fnd"),
           comment = c(ROR = "03wc8by49"))
  )
Description: Provides a scrolling chat interface with multiline input,
    suitable for creating chatbot apps based on Large Language Models
    (LLMs). Designed to work particularly well with the 'ellmer' R package
    for calling LLMs.
License: MIT + file LICENSE
URL: https://posit-dev.github.io/shinychat/r/,
        https://github.com/posit-dev/shinychat
BugReports: https://github.com/posit-dev/shinychat/issues
Imports: base64enc, bslib, cli, coro, ellmer (>= 0.4.1), fastmap,
        htmltools, jsonlite, lifecycle, promises (>= 1.3.2), rlang (>=
        1.1.0), S7, shiny (>= 1.10.0)
Suggests: covr, knitr, later, R6, testthat (>= 3.0.0), withr
Config/Needs/website: tidyverse/tidytemplate, rmarkdown, weathR, gt,
        glue, bsicons
Config/roxygen2/version: 8.0.0
Config/testthat/edition: 3
Encoding: UTF-8
NeedsCompilation: no
Packaged: 2026-06-01 16:51:50 UTC; cpsievert
Author: Joe Cheng [aut],
  Carson Sievert [aut],
  Garrick Aden-Buie [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-7111-0077>),
  Barret Schloerke [aut] (ORCID: <https://orcid.org/0000-0001-9986-114X>),
  Posit Software, PBC [cph, fnd] (ROR: <https://ror.org/03wc8by49>)
Maintainer: Garrick Aden-Buie <garrick@adenbuie.com>
Repository: CRAN
Date/Publication: 2026-06-01 19:10:02 UTC
Built: R 4.4.3; ; 2026-06-01 20:08:20 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(as.tags,shinychat_tool_card)
S3method(format,shinychat_tool_card)
S3method(knitr::knit_print,shinychat_tool_card)
S3method(print,shinychat_tool_card)
export(chat_app)
export(chat_append)
export(chat_append_message)
export(chat_clear)
export(chat_greeting)
export(chat_mod_server)
export(chat_mod_ui)
export(chat_restore)
export(chat_set_greeting)
export(chat_ui)
export(contents_shinychat)
export(markdown_stream)
export(output_markdown_stream)
export(update_chat_user_input)
if (getRversion() < "4.3.0") importFrom("S7", "@")
import(S7)
import(rlang)
importFrom(coro,async)
importFrom(htmltools,HTML)
importFrom(htmltools,as.tags)
importFrom(htmltools,css)
importFrom(htmltools,tag)
importFrom(lifecycle,deprecated)
importFrom(shiny,getDefaultReactiveDomain)

# shinychat

<a href="https://posit-dev.github.io/shinychat/py"><img src="https://posit-dev.github.io/shinychat/logo.svg" align="right" height="138" alt="shinychat for Python website" /></a>

Chat UI component for [Shiny for Python](https://shiny.posit.co/py/).

**Note:** shinychat is automatically installed with Shiny for Python and available as `shiny.ui.Chat` and `shiny.express.ui.Chat`. For complete instructions about creating chatbots please see the [Shiny for Python documentation](https://shiny.posit.co/py/docs/genai-chatbots.html).

## Installation

You can install shinychat from PyPI with:

```bash
uv pip install shinychat
```

Or, install the development version of shinychat from [GitHub](https://github.com/posit-dev/shinychat) with:

```bash
uv pip install git+https://github.com/posit-dev/shinychat.git
```

## Example

```r
from shiny.express import render, ui
from shinychat.express import Chat

# Set some Shiny page options
ui.page_opts(title="Hello Chat")

# Create a chat component, with an initial message
chat = Chat(
    id="chat",
    messages=[
        {"content": "Hello! How can I help you today?", "role": "assistant"},
    ],
)

# Display the chat
chat.ui()


# Define a callback to run when the user submits a message
@chat.on_user_submit
async def handle_user_input(user_input: str):
    await chat.append_message(f"You said: {user_input}")


"Message state:"


@render.code
def message_state():
    return str(chat.messages())
```

# shinychat <a href="https://posit-dev.github.io/shinychat/r/"><img src="man/figures/logo.svg" align="right" height="138" alt="shinychat for R website" /></a>

<!-- badges: start -->
[![R-CMD-check](https://github.com/posit-dev/shinychat/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/posit-dev/shinychat/actions/workflows/R-CMD-check.yaml)
[![CRAN status](https://www.r-pkg.org/badges/version/shinychat)](https://CRAN.R-project.org/package=shinychat)
[![shinychat status badge](https://posit-dev.r-universe.dev/shinychat/badges/version)](https://posit-dev.r-universe.dev/shinychat)
<!-- badges: end -->

**shinychat** provides a [Shiny](https://shiny.posit.co/) toolkit for building generative AI applications like chatbots and [streaming content](https://posit-dev.github.io/shinychat/r/reference/markdown_stream.html). It's designed to work alongside the [ellmer](https://ellmer.tidyverse.org/) package, which handles response generation.

## Installation

You can install shinychat from CRAN with:

``` r
install.packages("shinychat")
```

Or, install the development version of shinychat from [GitHub](https://github.com/) with:

``` r
# install.packages("pak")
pak::pak("posit-dev/shinychat/pkg-r")
```

## Example

To run this example, you'll first need to create an OpenAI API key, and set it in your environment as `OPENAI_API_KEY`.

You'll also need to install the [ellmer](https://ellmer.tidyverse.org/) package (with `install.packages("ellmer")`).

```r
library(shiny)
library(shinychat)

ui <- bslib::page_fillable(
  chat_ui(
    id = "chat",
    messages = "**Hello!** How can I help you today?"
  ),
  fillable_mobile = TRUE
)

server <- function(input, output, session) {
  chat <-
    ellmer::chat_openai(
      system_prompt = "Respond to the user as succinctly as possible."
    )

  observeEvent(input$chat_user_input, {
    stream <- chat$stream_async(input$chat_user_input)
    chat_append("chat", stream)
  })
}

shinyApp(ui, server)
```

<img src="man/figures/app.png" alt="Screenshot of the resulting app." style="width: 100%"/>

## Next steps

Ready to start building a chatbot with shinychat? See [Get Started](https://posit-dev.github.io/shinychat/r/articles/get-started.html) to learn more.

#' Split tag content around elements with data-shinychat-react
#'
#' Elements WITH the attribute are emitted bare.
#' Consecutive elements WITHOUT the attribute are grouped into
#' <shinychat-html> wrappers.
#'
#' @param content A tag, tagList, or other HTML content.
#' @return A list of tag children ready to be serialized.
#' @noRd
split_html_islands <- function(content) {
  # Convert to tags so custom classes (e.g., shinychat_tool_card)
  # resolve their data-shinychat-react attribute
  content <- htmltools::as.tags(content)

  if (inherits(content, "shiny.tag")) {
    if (has_react_attr(content)) {
      return(list(content))
    }
    return(list(htmltools::tag("shinychat-raw-html", list(content))))
  }

  if (inherits(content, "shiny.tag.list")) {
    children <- as.list(content)
  } else {
    return(list(htmltools::tag("shinychat-raw-html", list(content))))
  }

  if (length(children) == 0) {
    return(list())
  }

  is_react <- vapply(children, has_react_attr, logical(1))
  group_id <- cumsum(c(TRUE, diff(is_react) != 0))
  groups <- split(children, group_id)

  result <- list()
  for (group in groups) {
    if (has_react_attr(group[[1]])) {
      result <- c(result, group)
    } else {
      result <- c(result, list(htmltools::tag("shinychat-raw-html", group)))
    }
  }
  result
}

has_react_attr <- function(child) {
  if (!inherits(child, "shiny.tag")) {
    return(FALSE)
  }
  !is.null(htmltools::tagGetAttribute(child, "data-shinychat-react"))
}

test_that("plain HTML gets a single shinychat-raw-html wrapper", {
  content <- htmltools::tagList(
    htmltools::div("hello"),
    htmltools::span("world")
  )
  result <- split_html_islands(content)
  rendered <- as.character(htmltools::tagList(result))
  expect_equal(
    length(grep("<shinychat-raw-html>", strsplit(rendered, "\n")[[1]])),
    1
  )
  expect_match(rendered, "<div>hello</div>")
  expect_match(rendered, "<span>world</span>")
})

test_that("react element is emitted bare", {
  content <- htmltools::tag(
    "shiny-tool-result",
    list(`data-shinychat-react` = NA, `request-id` = "abc")
  )
  result <- split_html_islands(content)
  rendered <- as.character(htmltools::tagList(result))
  expect_no_match(rendered, "shinychat-raw-html")
  expect_match(rendered, "shiny-tool-result")
})

test_that("mixed content splits around react elements", {
  content <- htmltools::tagList(
    htmltools::div("before"),
    htmltools::tag(
      "shiny-tool-result",
      list(`data-shinychat-react` = NA, `request-id` = "abc")
    ),
    htmltools::div("after")
  )
  result <- split_html_islands(content)
  rendered <- as.character(htmltools::tagList(result))
  matches <- gregexpr("<shinychat-raw-html>", rendered)[[1]]
  expect_equal(sum(matches > 0), 2)
  expect_match(rendered, "shiny-tool-result")
})

test_that("adjacent react elements produce no empty islands", {
  content <- htmltools::tagList(
    htmltools::tag("shiny-tool-request", list(`data-shinychat-react` = NA)),
    htmltools::tag("shiny-tool-result", list(`data-shinychat-react` = NA))
  )
  result <- split_html_islands(content)
  rendered <- as.character(htmltools::tagList(result))
  expect_no_match(rendered, "shinychat-raw-html")
})

test_that("single tag without react attr gets wrapped", {
  content <- htmltools::div("hello")
  result <- split_html_islands(content)
  rendered <- as.character(htmltools::tagList(result))
  expect_match(rendered, "shinychat-raw-html")
})

test_that("empty tagList returns empty list without error", {
  result <- split_html_islands(htmltools::tagList())
  expect_equal(result, list())
})

#' Open a live chat application in the browser
#'
#' @description
#' Create a simple Shiny app for live chatting using an [ellmer::Chat] object.
#' Note that these functions will mutate the input `client` object as
#' you chat because your turns will be appended to the history.
#'
#' The app created by `chat_app()` is suitable for interactive use by a single
#' user. For multi-user Shiny apps, use the Shiny module chat functions --
#' `chat_mod_ui()` and `chat_mod_server()` -- and be sure to create a new chat
#' client for each user session.
#'
#' @examples
#' \dontrun{
#' # Interactive in the console ----
#' client <- ellmer::chat_anthropic()
#' chat_app(client)
#'
#' # Inside a Shiny app ----
#' library(shiny)
#' library(bslib)
#' library(shinychat)
#'
#' ui <- page_fillable(
#'   titlePanel("shinychat example"),
#'
#'   layout_columns(
#'     card(
#'       card_header("Chat with Claude"),
#'       chat_mod_ui(
#'         "claude",
#'         messages = list(
#'           "Hi! Use this chat interface to chat with Anthropic's `claude-3-5-sonnet`."
#'         )
#'       )
#'     ),
#'     card(
#'       card_header("Chat with ChatGPT"),
#'       chat_mod_ui(
#'         "openai",
#'         messages = list(
#'           "Hi! Use this chat interface to chat with OpenAI's `gpt-4o`."
#'         )
#'       )
#'     )
#'   )
#' )
#'
#' server <- function(input, output, session) {
#'   claude <- ellmer::chat_anthropic(model = "claude-3-5-sonnet-latest") # Requires ANTHROPIC_API_KEY
#'   openai <- ellmer::chat_openai(model = "gpt-4o") # Requires OPENAI_API_KEY
#'
#'   chat_mod_server("claude", claude)
#'   chat_mod_server("openai", openai)
#' }
#'
#' shinyApp(ui, server)
#' }
#'
#' @param client A chat object created by \pkg{ellmer}, e.g.
#'   [ellmer::chat_openai()] and friends. This argument is deprecated in
#'   `chat_mod_ui()` because the client state is now managed by
#'   `chat_mod_server()`.
#' @param ... In `chat_app()`, additional arguments are passed to
#'   [shiny::shinyApp()]. In `chat_mod_ui()`, additional arguments are passed to
#'   [chat_ui()].
#' @param bookmark_store The bookmarking store to use for the app. Passed to
#'   `enable_bookmarking` in [shiny::shinyApp()]. Defaults to `"url"`, which
#'   uses the URL to store the chat state. URL-based bookmarking is limited in
#'   size; use `"server"` to store the state on the server side without size
#'   limitations; or disable bookmarking by setting this to `"disable"`.
#'
#' @returns
#'   * `chat_app()` returns a [shiny::shinyApp()] object.
#'   * `chat_mod_ui()` returns the UI for a shinychat module.
#'   * `chat_mod_server()` includes the shinychat module server logic, and
#'     returns an environment containing:
#'
#'     * `last_input`: A reactive value containing the last user input.
#'     * `last_turn`: A reactive value containing the last assistant turn.
#'     * `update_user_input()`: A function to update the chat input or submit a
#'       new user input. Takes the same arguments as [update_chat_user_input()],
#'       except for `id` and `session`, which are supplied automatically.
#'     * `append()`: A function to append a new message to the chat UI. Takes
#'       the same arguments as [chat_append()], except for `id` and `session`,
#'       which are supplied automatically.
#'     * `clear()`: A function to clear the chat history and the chat UI.
#'       `clear()` takes an optional list of `messages` used to initialize the
#'       chat after clearing. `messages` should be a list of messages, where
#'       each message is a list with `role` and `content` fields. The
#'       `client_history` argument controls how the chat client's history is
#'       updated after clearing. It can be one of: `"clear"` the chat history;
#'       `"set"` the chat history to `messages`; `"append"` `messages` to the
#'       existing chat history; or `"keep"` the existing chat history.
#'     * `set_greeting()`: A function to set, stream, or clear the chat
#'       greeting. Pass a [chat_greeting()] object, a plain string, or
#'       `NULL` to clear. Streaming greetings run inside an
#'       [shiny::ExtendedTask] so the session stays responsive; if called
#'       while a greeting is already streaming, the new greeting is queued.
#'       If the greeting has already been dismissed, calling `set_greeting()`
#'       updates the content but does not make it visible again; call
#'       `clear(greeting = TRUE)` first to show a new greeting after dismissal.
#'     * `status`: A reactive value indicating the current chat interaction
#'       state. Returns `"idle"` when no response is in progress, or
#'       `"streaming"` while a response is actively being received.
#'     * `client`: The current chat client object (an active binding that
#'       always reflects the latest client, even after `set_client()`
#'       is called).
#'     * `set_client(new_client, sync = TRUE)`: Replace the chat client used by
#'       the module. When `sync` is `TRUE` (the default), the new client
#'       inherits conversation turns, system prompt, and tools from the previous
#'       client so the conversation continues seamlessly. Set `sync = FALSE` to
#'       use the new client as-is. If a response is currently streaming, the
#'       swap is deferred until the stream completes. If called multiple times
#'       while streaming, only the most recent new client is used.
#'
#' @describeIn chat_app A simple Shiny app for live chatting. Note that this
#'   app is suitable for interactive use by a single user; do not use
#'   `chat_app()` in a multi-user Shiny app context.
#' @export
chat_app <- function(client, ..., bookmark_store = "url") {
  check_ellmer_chat(client)

  ui <- function(req) {
    bslib::page_fillable(
      chat_mod_ui("chat", height = "100%"),
      shiny::actionButton(
        "close_btn",
        label = "",
        class = "btn-close",
        style = "position: fixed; top: 6px; right: 6px;"
      )
    )
  }

  server <- function(input, output, session) {
    chat_mod_server("chat", client)

    shiny::observeEvent(input$close_btn, label = "on_close_btn", {
      shiny::stopApp()
    })
  }

  shiny::shinyApp(ui, server, ..., enableBookmarking = bookmark_store)
}

check_ellmer_chat <- function(client) {
  if (!inherits(client, "Chat")) {
    abort("`client` must be an `ellmer::Chat` object.")
  }
}

#' @describeIn chat_app A simple chat app module UI.
#' @param id The chat module ID.
#' @param messages Initial messages shown in the chat, used only when `client`
#'   (in `chat_mod_ui()`) doesn't already contain turns. Passed to `messages`
#'   in [chat_ui()].
#' @export
chat_mod_ui <- function(
  id,
  ...,
  client = deprecated(),
  messages = NULL
) {
  if (lifecycle::is_present(client)) {
    lifecycle::deprecate_warn(
      "0.3.0",
      "chat_mod_ui(client = )",
      "chat_mod_server(client = )"
    )
  }

  chat_ui(
    shiny::NS(id, "chat"),
    messages = messages,
    enable_cancel = TRUE,
    ...
  )
}

#' @describeIn chat_app A simple chat app module server.
#' @inheritParams chat_restore
#' @param greeting Optional greeting to set when the module initializes.
#'   Accepts a static value (string, [htmltools::HTML()], [htmltools::tagList()],
#'   or [chat_greeting()]) or a **function** that generates the greeting
#'   dynamically. See the **Greeting** section below for details.
#'
#' @section Greeting:
#'
#' When `greeting` is a **function**, the module calls it each time the
#' `greeting_requested` event fires — on first view when the chat is empty,
#' and again after `clear(greeting = TRUE)`. The function should return a
#' [chat_greeting()] (typically wrapping a stream). Static values (strings,
#' [chat_greeting()] objects) are set once at init and do not regenerate.
#'
#' The module detects **named arguments** in the greeting function to decide
#' what to pass. Currently the only recognized argument is `client`.
#'
#' **`function(client)`** (recommended). The module clones the `client`
#' passed to `chat_mod_server()`, wipes its turn history, and passes the
#' fresh clone as `client`. This avoids manually creating and configuring
#' a separate client:
#'
#' ```r
#' chat_mod_server("chat", client, greeting = function(client) {
#'   stream <- client$stream_async("Generate a short welcome message.")
#'   chat_greeting(stream)
#' })
#' ```
#'
#' **`function()`** (zero arguments). You create and manage your own client:
#'
#' ```r
#' chat_mod_server("chat", client, greeting = function() {
#'   greeter <- ellmer::chat_openai(model = "gpt-4o")
#'   stream <- greeter$stream_async("Generate a short welcome message.")
#'   chat_greeting(stream)
#' })
#' ```
#'
#' **Static value.** Set once; does not regenerate after `clear()`:
#'
#' ```r
#' chat_mod_server("chat", client, greeting = "## Welcome!\n\nHow can I help?")
#' ```
#'
#' The returned `set_greeting()` helper is available for cases where you need
#' to set a greeting outside the greeting lifecycle.
#'
#' @export
chat_mod_server <- function(
  id,
  client,
  greeting = NULL,
  bookmark_on_input = TRUE,
  bookmark_on_response = TRUE
) {
  check_ellmer_chat(client)

  append_stream_task <- shiny::ExtendedTask$new(
    function(client, ui_id, user_input, controller = NULL) {
      stream <- client$stream_async(
        user_input,
        stream = "content",
        controller = controller
      )

      p <- promises::promise_resolve(stream)
      promises::then(p, function(stream) {
        chat_append(ui_id, stream)
      })
    }
  )

  greeting_stream_task <- shiny::ExtendedTask$new(
    function(ui_id, greeting, session) {
      result <- chat_set_greeting(ui_id, greeting, session = session)
      if (is.null(result)) {
        promises::promise_resolve(NULL)
      } else {
        result
      }
    }
  )

  shiny::moduleServer(id, function(input, output, session) {
    cancel_bookmarks <- chat_restore(
      "chat",
      client,
      session = session,
      bookmark_on_input = bookmark_on_input,
      bookmark_on_response = bookmark_on_response
    )

    last_turn <- shiny::reactiveVal(NULL, label = "last_turn")
    last_input <- shiny::reactiveVal(NULL, label = "last_input")
    pending_swap <- shiny::reactiveVal(NULL, label = "pending_swap")
    ctrl <- ellmer::stream_controller()

    swap_client <- function(new_client, sync) {
      if (sync) {
        new_client$set_turns(client$get_turns())
        new_client$set_system_prompt(client$get_system_prompt())
        new_client$set_tools(client$get_tools())
      }
      client <<- new_client
      cancel_bookmarks()
      cancel_bookmarks <<- chat_restore(
        "chat",
        client,
        session = session,
        bookmark_on_input = bookmark_on_input,
        bookmark_on_response = bookmark_on_response,
        restore_ui = FALSE
      )
      invisible()
    }

    set_client <- function(new_client, sync = TRUE) {
      check_ellmer_chat(new_client)

      if (append_stream_task$status() == "running") {
        pending_swap(list(client = new_client, sync = sync))
        return(invisible())
      }

      swap_client(new_client, sync)
    }

    shiny::observeEvent(input$chat_user_input, label = "on_chat_user_input", {
      last_input(input$chat_user_input)
      append_stream_task$invoke(
        client,
        "chat",
        input$chat_user_input,
        controller = ctrl
      )
    })

    shiny::observeEvent(input$chat_cancel, label = "on_chat_cancel", {
      ctrl$cancel()
    })

    shiny::observe(label = "on_stream_complete", {
      status <- append_stream_task$status()
      swap <- pending_swap()

      if (status == "success") {
        last_turn(client$last_turn())
      }

      if (!is.null(swap) && status != "running") {
        pending_swap(NULL)
        swap_client(swap$client, swap$sync)
      }
    })

    chat_update_user_input <- function(
      value = NULL,
      ...,
      placeholder = NULL,
      submit = FALSE,
      focus = FALSE
    ) {
      update_chat_user_input(
        "chat",
        value = value,
        placeholder = placeholder,
        submit = submit,
        focus = focus,
        ...,
        session = session
      )
    }

    chat_append_mod <- function(response, role = "assistant", icon = NULL) {
      chat_append("chat", response, role = role, icon = icon, session = session)
    }

    set_greeting_mod <- function(greeting) {
      greeting_stream_task$invoke("chat", greeting, session)
    }

    if (is.function(greeting)) {
      greeting_fmls <- names(formals(greeting))

      shiny::observeEvent(
        input$chat_greeting_requested,
        label = "on_greeting_requested",
        {
          args <- list()
          if ("client" %in% greeting_fmls) {
            greeter <- client$clone()
            greeter$set_turns(list())
            args$client <- greeter
          }
          greeting_stream_task$invoke(
            "chat",
            do.call(greeting, args),
            session
          )
        }
      )
    } else if (!is.null(greeting)) {
      set_greeting_mod(greeting)
    }

    client_clear <- function(
      messages = NULL,
      greeting = FALSE,
      client_history = c("clear", "set", "append", "keep")
    ) {
      client_history <- arg_match(client_history)

      if (!is.null(messages)) {
        if (rlang::is_string(messages)) {
          # Promote strings to single assistant message
          messages <- list(list(role = "assistant", content = messages))
        }
        if (!rlang::is_list(messages)) {
          cli::cli_abort(
            "{.var messages} must be a list of messages, and each message must be a list with {.field role} and {.field content}."
          )
        }
        if (length(intersect(c("role", "content"), names(messages))) == 2) {
          # Catch the single-message case and promote it to a list of messages
          messages <- list(messages)
        }
      }

      chat_clear("chat", greeting = greeting, session = session)
      if (!is.null(messages)) {
        for (msg in messages) {
          chat_append("chat", msg$content, role = msg$role, session = session)
        }
      }

      if (client_history == "clear") {
        client$set_turns(list())
      } else if (client_history == "set") {
        client$set_turns(as_ellmer_turns(messages))
      } else if (client_history == "append") {
        turns <- client$get_turns()
        turns <- c(turns, as_ellmer_turns(messages))
        client$set_turns(turns)
      }

      last_turn(NULL)
      last_input(NULL)
    }

    ret <- new.env(parent = emptyenv())
    ret$last_turn <- shiny::reactive(last_turn(), label = "mod_last_turn")
    ret$last_input <- shiny::reactive(last_input(), label = "mod_last_input")
    ret$status <- shiny::reactive(label = "mod_status", {
      if (append_stream_task$status() == "running") "streaming" else "idle"
    })
    makeActiveBinding("client", function() client, ret)
    ret$append <- chat_append_mod
    ret$update_user_input <- chat_update_user_input
    ret$clear <- client_clear
    ret$set_greeting <- set_greeting_mod
    ret$set_client <- set_client
    lockEnvironment(ret)
    ret
  })
}

render_tags <- function(ui) {
  res <- htmltools::renderTags(ui)

  deps <- lapply(res$dependencies, function(x) {
    x$version <- NULL
    x$src <- NULL
    unclass(x[!vapply(x, is.null, logical(1))])
  })

  list(
    deps = jsonlite::toJSON(deps, auto_unbox = TRUE),
    html = res$html
  )
}

as_ellmer_turns <- function(messages) {
  if (is.null(messages) || length(messages) == 0) {
    return(list())
  }

  map(messages, function(msg) {
    ellmer::Turn(
      role = msg$role,
      contents = list(ellmer::ContentText(msg$content))
    )
  })
}

.onLoad <- function(libname, pkgname) {
  rlang::run_on_load()
  S7::methods_register()
}

as_generator <- function(x) {
  if (inherits(x, "coro_generator_instance")) {
    x
  } else {
    coro::gen(yield(x))
  }
}

pre_process_ui <- function(ui) {
  htmltools::tagList(!!!split_html_islands(ui))
}

process_ui <- function(ui, session) {
  process_deps <- asNamespace("shiny")[["processDeps"]]
  if (!is.function(process_deps)) {
    stop(
      "Expected processDeps() function to exist in Shiny. Please report this issue."
    )
  }

  # Render UI to html and register dependencies with the session
  res <- with_current_theme({
    process_deps(ui, session)
  })

  # Remove html_dependency class so jsonlite can handle it
  res[["deps"]] <- lapply(res[["deps"]], unclass)
  res
}

# Compile HTMLDependency()s against the current/default theme
# (that is, compile Sass with Bootstrap Sass headers)
with_current_theme <- function(expr) {
  theme <- bslib::bs_current_theme() %||% bslib::bs_theme()
  old_theme <- bslib::bs_global_set(theme)
  on.exit(bslib::bs_global_set(old_theme), add = TRUE)
  force(expr)
}

tag_require <- function(tag, version = 5, caller = "") {
  tag_req <- asNamespace("bslib")[["tag_require"]]
  if (!is.function(tag_req)) {
    stop(
      "Expected tag_require() function to exist in bslib. Please report this issue."
    )
  }
  tag_req(tag, version, caller)
}

check_active_session <- function(session = shiny::getDefaultReactiveDomain()) {
  if (is.null(session)) {
    rlang::abort(
      "An active Shiny session is required.",
      call = rlang::caller_env()
    )
  }
}

resolve_id <- function(id, session = shiny::getDefaultReactiveDomain()) {
  if (is.null(session)) {
    return(id)
  }
  session$ns(id)
}

send_chat_action <- function(id, action, html_deps = NULL, session) {
  envelope <- list(
    id = resolve_id(id, session),
    action = action
  )
  if (!is.null(html_deps)) {
    envelope$html_deps <- html_deps
  }
  session$sendCustomMessage("shinyChatMessage", envelope)
}

#' Add Shiny bookmarking for shinychat
#'
#' @description
#' Adds Shiny bookmarking hooks to save and restore the \pkg{ellmer} chat
#' `client`. Also restores chat messages from the history in the `client`.
#'
#' If either `bookmark_on_input` or `bookmark_on_response` is `TRUE`, the Shiny
#' App's bookmark will be automatically updated without showing a modal to the
#' user.
#'
#' Note: Only the `client`'s chat state is saved/restored in the bookmark. If
#' the `client`'s state doesn't properly capture the chat's UI (i.e., a
#' transformation is applied in-between receiving and displaying the message),
#' then you may need to implement your own `session$onRestore()` (and possibly
#' `session$onBookmark`) handler to restore any additional state.
#'
#' To avoid restoring chat history from the `client`, you can ensure that the
#' history is empty by calling `client$set_turns(list())` before passing the
#' client to `chat_restore()`.
#'
#' @param id The ID of the chat element
#' @param client The \pkg{ellmer} LLM chat client.
#' @param ... Used for future parameter expansion.
#' @param bookmark_on_input A logical value determines if the bookmark should be updated when the user submits a message. Default is `TRUE`.
#' @param bookmark_on_response A logical value determines if the bookmark should be updated when the response stream completes. Default is `TRUE`.
#' @param restore_ui Whether to render the client's existing turns into the
#'   chat UI on registration. Default is `TRUE`. Set to `FALSE` when
#'   re-registering bookmarks after a client swap (where the UI already reflects
#'   the conversation).
#' @param session The Shiny session object
#' @returns Invisibly returns a function that, when called, cancels all
#'   bookmark registrations made by this call. This is useful when swapping
#'   the chat client: cancel the previous bookmarks, then call
#'   `chat_restore()` again with the new client.
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#' library(shinychat)
#'
#' ui <- function(request) {
#'   page_fillable(
#'     chat_ui("chat", fill = TRUE)
#'   )
#' }
#'
#' server <- function(input, output, session) {
#'   chat_client <- ellmer::chat_ollama(
#'     system_prompt = "Important: Always respond in a limerick",
#'     model = "qwen2.5-coder:1.5b",
#'     echo = TRUE
#'   )
#'   # Update bookmark to chat on user submission and completed response
#'   chat_restore("chat", chat_client)
#'
#'   observeEvent(input$chat_user_input, {
#'     stream <- chat_client$stream_async(input$chat_user_input)
#'     chat_append("chat", stream)
#'   })
#' }
#'
#' # Enable bookmarking!
#' shinyApp(ui, server, enableBookmarking = "server")
#' @export
chat_restore <- function(
  id,
  client,
  ...,
  bookmark_on_input = TRUE,
  bookmark_on_response = TRUE,
  restore_ui = TRUE,
  session = getDefaultReactiveDomain()
) {
  rlang::check_dots_empty()
  stopifnot(is.character(id) && length(id) == 1)

  rlang::check_installed("ellmer")
  if (!(inherits(client, "R6") && inherits(client, "Chat"))) {
    rlang::abort(
      "`client` must be an `ellmer::Chat()` object. If you would like to have {shinychat} support your own package, please submit a GitHub Issue at https://github.com/posit-dev/shinychat"
    )
  }
  bookmark_on_input <- rlang::is_true(bookmark_on_input)
  bookmark_on_response <- rlang::is_true(bookmark_on_response)

  if (is.null(session)) {
    rlang::abort(
      "A `session` must be provided. Be sure to call `chat_restore()` where a session context is available."
    )
  }

  # Exclude works with bookmark names
  excluded_names <- session$getBookmarkExclude()
  id_user_input <- paste0(id, "_user_input")
  if (!(id_user_input %in% excluded_names)) {
    session$setBookmarkExclude(c(excluded_names, id_user_input))
  }

  # Save
  cancel_on_bookmark_client <-
    session$onBookmark(function(state) {
      if (id %in% names(state$values)) {
        rlang::abort(
          paste0(
            "Bookmark value with id (`\"",
            id,
            "\"`)) already exists. Please remove it or use a different id."
          )
        )
      }

      client_state <- client_get_state(client)

      state$values[[id]] <- client_state
    })

  cancel_set_ui <- NULL
  if (restore_ui) {
    cancel_set_ui <- shiny::observe(label = "set_ui", {
      client_set_ui(client, id = id)
      cancel_set_ui$destroy()
    })
  }

  # Restore
  cancel_on_restore_client <-
    session$onRestore(function(state) {
      client_state <- state$values[[id]]
      if (is.null(client_state)) {
        return()
      }

      if (!is.null(cancel_set_ui)) {
        cancel_set_ui$destroy()
      }
      client_set_state(client, client_state)

      # Set the UI
      shiny::withReactiveDomain(session, {
        client_set_ui(client, id = id)
      })
    })

  # Update URL
  cancel_bookmark_on_input <-
    if (bookmark_on_input) {
      shiny::observeEvent(
        session$input[[id_user_input]],
        label = "on_user_submit_do_bookmark",
        {
          # On user submit
          session$doBookmark()
        }
      )
    } else {
      NULL
    }

  # Enable (or disable) session auto bookmarking if at least one chat wants it
  set_session_bookmark_on_response(
    session,
    id,
    enable = bookmark_on_response
  )

  cancel_update_bookmark <- NULL
  if (bookmark_on_input || bookmark_on_response) {
    cancel_update_bookmark <-
      shiny::withReactiveDomain(session$rootScope(), {
        # Update the query string when bookmarked
        shiny::onBookmarked(function(url) {
          shiny::updateQueryString(url)
        })
      })
  }

  # Return a single cancel callback that tears down all registrations
  cancel_all <- function() {
    # session$onBookmark() and shiny::onBookmarked() return cancel functions
    if (!is.null(cancel_on_bookmark_client)) {
      cancel_on_bookmark_client()
    }
    if (!is.null(cancel_on_restore_client)) {
      cancel_on_restore_client()
    }
    if (!is.null(cancel_update_bookmark)) {
      cancel_update_bookmark()
    }
    # observeEvent() returns an Observer with $destroy()
    if (!is.null(cancel_bookmark_on_input)) cancel_bookmark_on_input$destroy()
  }

  invisible(cancel_all)
}

# Method currently hooked into `chat_append_stream()` and `markdown_stream()`
# When the incoming stream ends, possibly update the URL given the `id`
chat_update_bookmark <- function(
  id,
  stream_promise,
  session = shiny::getDefaultReactiveDomain()
) {
  if (!has_session_bookmark_on_response(session, id)) {
    # No auto bookmark set. Return early!
    return(stream_promise)
  }

  # Bookmark has been flagged for `id`.
  # When the stream ends, update the URL.
  prom <-
    promises::then(stream_promise, function(stream) {
      # Force a bookmark update when the stream ends!
      shiny::isolate(session$doBookmark())
    })

  return(prom)
}

# These methods exist to set flags within the session.
# These flags will determine if the session should be bookmarked when a response has completed.
# `chat_update_bookmark()` will check if the flag is set and update the URL if it is.
ON_RESPONSE_KEY <- ".bookmark-on-response"
has_session_bookmark_on_response <- function(session, id) {
  has_session_chat_bookmark_info(
    session,
    paste0(id, ON_RESPONSE_KEY)
  )
}
set_session_bookmark_on_response <- function(session, id, enable) {
  set_session_chat_bookmark_info(
    session,
    paste0(id, ON_RESPONSE_KEY),
    value = if (enable) TRUE else NULL
  )
}

has_session_chat_bookmark_info <- function(session, id) {
  return(!is.null(get_session_chat_bookmark_info(session, id)))
}
get_session_chat_bookmark_info <- function(session, id) {
  if (is.null(session)) {
    return(NULL)
  }

  info <- session$userData$shinychat
  key <- session$ns(id)
  return(info[[key]])
}
set_session_chat_bookmark_info <- function(session, id, value) {
  if (is.null(session)) {
    return(NULL)
  }

  if (is.null(session$userData$shinychat)) {
    session$userData$shinychat <- list()
  }
  session$userData$shinychat[[session$ns(id)]] <- value

  invisible(session)
}

library(shiny)
library(ellmer)
library(shinychat)

rlang::check_installed("ellmer", version = "0.3.0")

PROVIDER_MODEL = Sys.getenv("TEST_PROVIDER_MODEL", "openai/gpt-4.1-nano")

TOOL_OPTS <- list(
  async = as.logical(Sys.getenv("TEST_TOOL_ASYNC", "TRUE")),
  with_intent = as.logical(Sys.getenv("TEST_TOOL_WITH_INTENT", "TRUE")),
  with_title = as.logical(Sys.getenv("TEST_TOOL_WITH_TITLE", "TRUE")),
  with_icon = as.logical(Sys.getenv("TEST_TOOL_WITH_ICON", "TRUE"))
)

ui <- bslib::page_fillable(
  shinychat::chat_mod_ui(
    "chat",
    messages = list(
      list(
        role = "assistant",
        content = htmltools::HTML(
          '<span class="suggestion submit">In three separate but parallel tool calls list the files in apps, data, docs</span>

<span class="suggestion submit">Write some basic R code that demonstrates how to use the tidyverse.</span>

<span class="suggestion submit">Brainstorm 10 ideas for a name for a package that creates interactive sparklines in tables.</span>'
        )
      )
    )
  ),
  actionButton("click", "Click me")
)

maybe_fail <- function() {
  if (sample(c(TRUE, FALSE), 1, prob = c(0.25, 0.75))) {
    stop("An error occurred while listing files.")
  }
}

tool_fun <- if (TOOL_OPTS$async) {
  # Use async function for tool call
  if (TOOL_OPTS$with_intent) {
    # Async function with intent
    coro::async(function(path, `_intent` = "") {
      coro::await(coro::async_sleep(runif(1, 1, 10)))
      maybe_fail()
      c("app.R", "data.csv")
    })
  } else {
    # Async function without intent
    coro::async(function(path) {
      coro::await(coro::async_sleep(runif(1, 1, 10)))
      maybe_fail()
      c("app.R", "data.csv")
    })
  }
} else {
  # Synchronous version
  if (TOOL_OPTS$with_intent) {
    # Synchronous function with intent
    function(path, `_intent` = "") {
      Sys.sleep(runif(1, 1, 3))
      maybe_fail()
      c("app.R", "data.csv")
    }
  } else {
    # Synchronous function without intent
    function(path) {
      Sys.sleep(runif(1, 1, 3))
      maybe_fail()
      c("app.R", "data.csv")
    }
  }
}

tool_args <- list(
  path = type_string("Path to the directory to list files"),
  `_intent` = if (TOOL_OPTS$with_intent) {
    type_string(
      "Reason for the request to explain the tool call to the user"
    )
  }
)

tool_annotations <- list(
  title = if (TOOL_OPTS$with_title) "List Files",
  description = "This tool lists files in the specified directory.",
  icon = if (TOOL_OPTS$with_icon) bsicons::bs_icon("folder2-open")
)

server <- function(input, output, session) {
  packaged_list_files_tool <- tool(
    tool_fun,
    "List files in the user's current directory. Always check again when asked.",
    arguments = purrr::compact(tool_args),
    name = "list_file",
    annotations = purrr::compact(tool_annotations)
  )

  client <- chat(PROVIDER_MODEL)
  client$register_tool(packaged_list_files_tool)

  chat_mod_server("chat", client)

  observeEvent(input$click, {
    updateActionButton(
      session,
      "click",
      label = sprintf("Clicked %d times", input$click)
    )
  })
}

shinyApp(ui, server)

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import rlang
#' @import S7
#' @importFrom coro async
#' @importFrom htmltools as.tags
#' @importFrom htmltools tag css HTML
#' @importFrom lifecycle deprecated
## usethis namespace: end
NULL

# enable usage of <S7_object>@name in package code
#' @rawNamespace if (getRversion() < "4.3.0") importFrom("S7", "@")
NULL

ignore_unused_imports <- function() {
  jsonlite::fromJSON
  fastmap::fastqueue
  ellmer::contents_html
}

release_bullets <- function() {
  c(
    "Check that shinychat js assets are up-to-date (`make js-build && make r-update-dist`)"
  )
}

sanitized_chat_error <- function(err) {
  needs_sanitized <-
    isTRUE(getOption("shiny.sanitize.errors")) &&
    !inherits(err, "shiny.custom.error")

  if (needs_sanitized) {
    "\n\n**An error occurred.** Please try again or contact the app author."
  } else {
    sprintf(
      "\n\n**An error occurred:**\n\n```\n%s\n```",
      strip_ansi(conditionMessage(err))
    )
  }
}

strip_ansi <- function(text) {
  # Matches codes like "\x1B[31;43m", "\x1B[1;3;4m"
  ansi_pattern <- "(\x1B|\x033)\\[[0-9;?=<>]*[@-~]"
  gsub(ansi_pattern, "", text)
}

shinychat_deps <- function() {
  htmltools::htmlDependency(
    "shinychat",
    utils::packageVersion("shinychat"),
    package = "shinychat",
    src = "lib/shiny",
    script = list(src = "shinychat.js", type = "module"),
    stylesheet = "shinychat.css"
  )
}

drop_nulls <- function(x) {
  x[!vapply(x, is.null, FUN.VALUE = logical(1))]
}

library(shiny)
library(bslib)
library(ellmer)
library(shinychat)
library(weathR)

Sys.setenv("SHINYCHAT_TOOL_DISPLAY" = "rich")

get_weather_forecast <- tool(
  function(lat, lon, location_name) {
    forecast_data <- point_tomorrow(lat, lon, short = FALSE)
    forecast_table <- gt::as_raw_html(gt::gt(forecast_data))

    ContentToolResult(
      forecast_data,
      extra = list(
        display = list(
          html = forecast_table,
          title = paste("Weather Forecast for", location_name),
          full_screen = TRUE
        )
      )
    )
  },
  name = "get_weather_forecast",
  description = "Get the weather forecast for a location.",
  arguments = list(
    lat = type_number("Latitude"),
    lon = type_number("Longitude"),
    location_name = type_string("Name of the location for display to the user")
  ),
  annotations = tool_annotations(
    title = "Weather Forecast",
    icon = bsicons::bs_icon("cloud-sun")
  )
)

ui <- function(req) {
  page_fillable(
    chat_mod_ui("chat")
  )
}

server <- function(input, output, session) {
  client <- ellmer::chat("openai/gpt-4.1-nano")
  # client <- ellmer::chat_ollama(model = "mistral-nemo")
  client$register_tool(get_weather_forecast)
  chat_mod_server("chat", client)
}

shinyApp(ui, server, enableBookmarking = "url")

library(shiny)
library(bslib)
library(ellmer)
library(shinychat)
library(weathR)

# This example shows how to use a custom tool result class. It extends the
# contents_shinychat() generic to compute the HTML table on the fly when we
# render the result in the chat interface. This allows the tool result object to
# be lighter-weight and to only hold the raw data and metadata, without needing
# to also pre-compute the HTML table.

WeatherToolResult <- S7::new_class(
  "WeatherToolResult",
  parent = ContentToolResult,
  properties = list(
    location_name = S7::class_character
  )
)

contents_shinychat <- S7::new_external_generic(
  "shinychat",
  "contents_shinychat",
  "contents"
)

S7::method(contents_shinychat, WeatherToolResult) <- function(content) {
  current <- content@value[1, ]

  bslib::value_box(
    title = content@location_name,
    value = current$skies,
    showcase = bsicons::bs_icon("cloud-sun"),
    full_screen = TRUE,
    sprintf(
      "%s°F (High: %s°F, Low: %s°F)",
      current$temp,
      max(content@value$temp),
      min(content@value$temp)
    )
  )
}

get_weather_forecast <- tool(
  function(lat, lon, location_name) {
    WeatherToolResult(
      point_tomorrow(lat, lon, short = FALSE),
      location_name = location_name
    )
  },
  name = "get_weather_forecast",
  description = "Get the weather forecast for a location.",
  arguments = list(
    lat = type_number("Latitude"),
    lon = type_number("Longitude"),
    location_name = type_string("Name of the location for display to the user")
  ),
  annotations = tool_annotations(
    title = "Weather Forecast",
    icon = bsicons::bs_icon("cloud-sun")
  )
)

ui <- function(req) {
  page_fillable(
    chat_mod_ui("chat")
  )
}

server <- function(input, output, session) {
  client <- ellmer::chat("openai/gpt-4.1-nano")
  # client <- ellmer::chat_ollama(model = "mistral-nemo")
  client$register_tool(get_weather_forecast)
  chat_mod_server("chat", client)
}

shinyApp(ui, server, enableBookmarking = "url")

client_get_state <- new_generic("client_get_state", "client")

client_set_state <- new_generic("client_get_state", "client")

client_set_ui <- new_generic(
  "client_set_ui",
  "client",
  function(client, ..., id) {
    S7::S7_dispatch()
  }
)

is_server_bookmarkstore <- function() {
  shiny::getShinyOption("bookmarkStore", "") == "server"
}
is_url_bookmarkstore <- function() {
  shiny::getShinyOption("bookmarkStore", "") == "url"
}

method(client_get_state, S7::new_S3_class(c("Chat", "R6"))) <-
  function(client) {
    # Do not record the client object itself. This would be a security leak.
    # Instead, save only the `turns` information
    recorded_turns <- lapply(
      client$get_turns(),
      ellmer::contents_record
    )

    if (is_url_bookmarkstore()) {
      recorded_turns <- lapply(
        recorded_turns,
        function(turn) {
          turn$props$json <- NULL
          turn
        }
      )
    }

    # Pre-serialize the contents so that when shiny:::toJSON() is called, it is stable.
    # jsonlite::toJSON() is not stable as it is a lossy serialization. In addition, jsonlite::fromJSON() (which shiny:::safeFromJSON() uses) is not stable as it tries to make everything a data.frame.
    #
    # * `jsonlite::serializeJSON()` is a stable transformation
    # * `jsonlite::unserializeJSON()` is a stable transformation
    state_json <- jsonlite::serializeJSON(recorded_turns)
    state_str <- base64enc::base64encode(memCompress(state_json, "gzip"))

    list(
      version = 1,
      state = state_str
    )
  }

method(client_set_state, S7::new_S3_class(c("Chat", "R6"))) <-
  function(client, state) {
    if (!is.list(state)) {
      rlang::abort(
        "Invalid state. Expected a list with a 'version' and 'state' element."
      )
    }
    if (state$version != 1) {
      rlang::abort(
        paste0("Invalid state version. Expected 1, got ", state$version)
      )
    }

    state_str <- state$state

    state_json <- memDecompress(
      base64enc::base64decode(state_str),
      asChar = TRUE
    )
    recorded_turns <- jsonlite::unserializeJSON(state_json)

    replayed_turns <- lapply(
      recorded_turns,
      ellmer::contents_replay,
      tools = client$get_tools()
    )

    client$set_turns(replayed_turns)
  }

method(client_set_ui, S7::new_S3_class(c("Chat", "R6"))) <-
  function(client, ..., id) {
    # TODO-future: Disable bookmarking when restoring. Leverage `tryCatch(finally={})`
    # TODO-barret-future; In shinychat, make this a single/internal custom message call to send all the messages at once (and then scroll)

    msgs <- contents_shinychat(client)
    lapply(msgs, function(msg_turn) {
      is_list <- is.list(msg_turn$content) &&
        !inherits(msg_turn$content, c("shiny.tag", "shiny.taglist"))

      if (is_list) {
        stream <- coro::generator(function() {
          for (x in msg_turn$content) {
            coro::yield(x)
          }
        })
        chat_append(id, stream(), msg_turn$role)
      } else {
        chat_append(id, msg_turn$content, role = msg_turn$role)
      }
    })
  }

# Used to avoid R CMD check NOTE about unused imports
`_ignore` <- function() {
  base64enc::base64encode
}

library(shiny)
library(bslib)
library(ellmer)
library(shinychat)
library(weathR)

get_weather_forecast <- tool(
  function(lat, lon, location_name) {
    forecast <- point_tomorrow(lat, lon, short = FALSE)

    icon <- if (any(forecast$temp > 70)) {
      bsicons::bs_icon("sun-fill")
    } else if (any(forecast$temp < 45)) {
      bsicons::bs_icon("snow")
    } else {
      bsicons::bs_icon("cloud-sun-fill")
    }

    ContentToolResult(
      forecast,
      extra = list(
        display = list(
          title = paste("Weather Forecast for", location_name),
          icon = icon
        )
      )
    )
  },
  name = "get_weather_forecast",
  description = "Get the weather forecast for a location.",
  arguments = list(
    lat = type_number("Latitude"),
    lon = type_number("Longitude"),
    location_name = type_string("Name of the location for display to the user")
  ),
  annotations = tool_annotations(
    title = "Weather Forecast",
    icon = bsicons::bs_icon("cloud-sun")
  )
)

ui <- function(req) {
  page_fillable(
    chat_mod_ui("chat")
  )
}

server <- function(input, output, session) {
  client <- ellmer::chat("openai/gpt-4.1-nano")
  # client <- ellmer::chat_ollama(model = "mistral-nemo")
  client$register_tool(get_weather_forecast)
  chat_mod_server("chat", client)
}

shinyApp(ui, server, enableBookmarking = "url")

local_shinychat_tool_display <- function(
  envvar = NULL,
  opt = NULL,
  env = parent.frame()
) {
  withr::local_envvar(SHINYCHAT_TOOL_DISPLAY = envvar, .local_envir = env)
  withr::local_options(shinychat.tool_display = opt, .local_envir = env)
}

with_shinychat_tool_display <- function(
  code,
  envvar = NULL,
  opt = NULL
) {
  local_shinychat_tool_display(envvar, opt)
  force(code)
}

# Helper function to create tool requests
new_tool_request <- function(
  id = "test-id",
  name = "test-tool",
  arguments = list(),
  tool = NULL
) {
  ellmer::ContentToolRequest(
    id = id,
    name = name,
    arguments = arguments,
    tool = tool %||% new_tool()
  )
}

# Helper function to create tool results
new_tool_result <- function(
  value = NULL,
  error = NULL,
  request = new_tool_request(),
  extra = list()
) {
  ellmer::ContentToolResult(
    value = value,
    error = error,
    request = request,
    extra = extra
  )
}

# Helper function to create a basic tool definition
new_tool <- function(name = NULL, description = NULL, annotations = list()) {
  ellmer::tool(
    function() NULL,
    name = name %||% "test_tool",
    description = description %||% "A test tool to test tooling",
    annotations = annotations
  )
}

test_that("opt_shinychat_tool_display handles options and environment variables", {
  withr::local_options(list(shinychat.tool_display = NULL))
  withr::local_envvar(list(SHINYCHAT_TOOL_DISPLAY = NULL))

  # Default behavior
  with_shinychat_tool_display({
    expect_equal(opt_shinychat_tool_display(), "rich")
  })

  # Option setting
  with_shinychat_tool_display(opt = "basic", {
    expect_equal(opt_shinychat_tool_display(), "basic")
  })

  # Environment variable
  with_shinychat_tool_display(envvar = "none", {
    expect_equal(opt_shinychat_tool_display(), "none")
  })

  # Option takes precedence over env var
  with_shinychat_tool_display(envvar = "none", opt = "basic", {
    expect_equal(opt_shinychat_tool_display(), "basic")
  })

  # Invalid values
  with_shinychat_tool_display(envvar = "invalid", {
    expect_snapshot(
      error = TRUE,
      opt_shinychat_tool_display()
    )
  })
  with_shinychat_tool_display(opt = "invalid", {
    expect_snapshot(
      error = TRUE,
      opt_shinychat_tool_display()
    )
  })
})

test_that("basic Content handling works", {
  ContentHTML <- S7::new_class(
    "ContentHTML",
    parent = ellmer::ContentText
  )
  S7::method(contents_shinychat, ContentHTML) <- function(content) {
    shiny::HTML(content@text)
  }

  ContentMarkdown <- S7::new_class(
    "ContentMarkdown",
    parent = ellmer::ContentText
  )
  S7::method(contents_shinychat, ContentMarkdown) <- function(content) {
    content@text
  }

  # Test HTML content
  html_content <- ContentHTML(HTML("<p>test</p>"))
  expect_equal(
    as.character(contents_shinychat(html_content)),
    "<p>test</p>"
  )

  # Test Markdown content
  md_content <- ContentMarkdown("**test**")
  expect_equal(contents_shinychat(md_content), "**test**")

  # Test Text content
  text_content <- ellmer::ContentText("test")
  expect_equal(contents_shinychat(text_content), "test")
})

test_that("ContentToolRequest returns NULL when display is disabled", {
  # Should return NULL when display is none
  with_shinychat_tool_display(opt = "none", {
    request <- new_tool_request()
    expect_null(contents_shinychat(request))
  })
})

test_that("ContentToolRequest rich display", {
  local_shinychat_tool_display(opt = "rich")

  request <- new_tool_request(
    id = "test-123",
    name = "weather",
    arguments = list(`_intent` = "Check weather", location = "NYC")
  )

  res <- contents_shinychat(request)
  expect_s3_class(res, "shinychat_tool_request")
  expect_equal(res$request_id, "test-123")
  expect_equal(res$tool_name, "weather")
  expect_equal(res$intent, "Check weather")
  expect_equal(
    jsonlite::fromJSON(res$arguments),
    list(`_intent` = "Check weather", location = "NYC")
  )

  res_tags <- as.tags(res)
  expect_equal(res_tags$name, "shiny-tool-request")
  expect_equal(res_tags$attribs$"request-id", "test-123")
  expect_equal(res_tags$attribs[["tool-name"]], "weather")
  expect_equal(res_tags$attribs$intent, "Check weather")
  expect_equal(
    jsonlite::fromJSON(res_tags$attribs$arguments),
    list(`_intent` = "Check weather", location = "NYC")
  )
})

test_that("ContentToolRequest handles tool annotations", {
  local_shinychat_tool_display(opt = "rich")

  tool <- new_tool(
    name = "weather",
    annotations = list(title = "Weather Tool")
  )
  request <- new_tool_request(tool = tool)
  res <- contents_shinychat(request)

  expect_s3_class(res, "shinychat_tool_request")
  expect_equal(res$tool_title, "Weather Tool")
})

test_that("ContentToolResult requires an associated `@request` property", {
  expect_snapshot(
    error = TRUE,
    contents_shinychat(new_tool_result(request = NULL))
  )
})

test_that("returns NULL for ContentToolResult when display is none", {
  local_shinychat_tool_display(opt = "none")

  base_request <- new_tool_request()
  result <- new_tool_result(request = base_request)

  expect_null(contents_shinychat(result))
})

test_that("simple ContentToolResult are displayed correctly", {
  local_shinychat_tool_display(opt = "rich")

  result <- new_tool_result(value = "Success!")
  res <- contents_shinychat(result)

  expect_s3_class(res, "shinychat_tool_result")
  expect_equal(res$request_id, result@request@id)
  expect_equal(res$tool_name, result@request@name)
  expect_equal(res$value, "Success!")
  expect_equal(res$value_type, "code")
  expect_equal(res$status, "success")
})

test_that("errors in ContentToolResult are displayed correctly", {
  local_shinychat_tool_display(opt = "rich")

  result <- new_tool_result(error = "Failed!")
  res <- contents_shinychat(result)

  expect_s3_class(res, "shinychat_tool_result")
  expect_equal(res$status, "error")
  expect_equal(res$value, "Failed!")
  expect_equal(res$value_type, "code")

  # basic and rich display are the same
  expect_equal(
    with_shinychat_tool_display(opt = "basic", contents_shinychat(result)),
    res
  )
})

test_that("ContentToolResult with custom text display", {
  local_shinychat_tool_display(opt = "rich")

  result <- new_tool_result(
    value = "success",
    extra = list(display = list(text = "Success!"))
  )

  expect_equal(
    tool_result_display(result),
    list(value = "Success!", value_type = "text")
  )

  res <- contents_shinychat(result)
  expect_s3_class(res, "shinychat_tool_result")
  expect_equal(res$request_id, result@request@id)
  expect_equal(res$tool_name, result@request@name)
  expect_equal(res$status, "success")
  expect_equal(res$value, "Success!")
  expect_equal(res$value_type, "text")
  expect_equal(res$show_request, NA)
  expect_null(res$expanded)

  res_tags <- as.tags(res)
  expect_s3_class(res_tags, "shiny.tag")
  expect_equal(res_tags$name, "shiny-tool-result")
  expect_equal(res_tags$attribs$status, "success")
  expect_equal(res_tags$attribs$value, "Success!")
  expect_equal(res_tags$attribs$"value-type", "text")
  expect_equal(res_tags$attribs[["show-request"]], NA)
  expect_null(res_tags$attribs$expanded)
})

test_that("ContentToolResult with additional display options from result", {
  local_shinychat_tool_display(opt = "rich")

  result <- new_tool_result(
    value = "test",
    extra = list(
      display = list(
        html = "<p>test</p>",
        show_request = FALSE,
        open = TRUE,
        title = "Custom Title"
      )
    )
  )
  res <- contents_shinychat(result)
  expect_s3_class(res, "shinychat_tool_result")
  expect_equal(res$value, "<p>test</p>")
  expect_equal(res$value_type, "html")
  expect_equal(res$show_request, NULL)
  expect_equal(res$expanded, NA)
  expect_equal(res$tool_title, "Custom Title")

  res_tags <- as.tags(res)
  expect_equal(res_tags$attribs$value, "<p>test</p>")
  expect_equal(res_tags$attribs$"value-type", "html")
  expect_equal(res_tags$attribs[["show-request"]], NULL)
  expect_equal(res_tags$attribs$expanded, NA)
  expect_equal(res_tags$attribs[["tool-title"]], "Custom Title")
})

test_that("ContentToolResult with HTML() title preserves markup", {
  local_shinychat_tool_display(opt = "rich")

  result <- new_tool_result(
    value = "test",
    extra = list(
      display = list(
        text = "test",
        title = HTML("Map of <i>Paris</i>")
      )
    )
  )
  res <- contents_shinychat(result)
  expect_s3_class(res$tool_title, "html")
  expect_equal(as.character(res$tool_title), "Map of <i>Paris</i>")

  # htmltools always escapes attribute values, but the browser decodes them,
  # so JS getAttribute() returns the original HTML string. The Playwright

  # test (test_html_title.py) verifies the end-to-end rendering.
  res_tags <- as.tags(res)
  expect_equal(res_tags$attribs[["tool-title"]], HTML("Map of <i>Paris</i>"))
})

test_that("ContentToolResult handles icon and dependencies from tool definition", {
  local_shinychat_tool_display(opt = "rich")

  icon_dep <- htmltools::htmlDependency(
    name = "test",
    version = "1.0",
    src = "."
  )

  tool <- new_tool(
    annotations = list(
      icon = htmltools::tags$i(class = "icon", icon_dep)
    )
  )
  result <- new_tool_result(
    value = "test",
    request = new_tool_request(tool = tool),
    extra = list(display = list(text = "test"))
  )

  res <- contents_shinychat(result)
  expect_s3_class(res, "shinychat_tool_result")
  expect_equal(res$icon, tool@annotations$icon)

  res_tags <- as.tags(res)
  expect_equal(format(res_tags$attribs$icon), '<i class="icon"></i>')
  expect_true(
    list(icon_dep) %in% htmltools::findDependencies(res_tags$children)
  )
})

test_that("ContentToolResult formats request_call correctly", {
  local_shinychat_tool_display(opt = "rich")

  result <- new_tool_result(
    value = "test",
    request = new_tool_request(
      name = "test",
      arguments = list(x = 1, y = "test")
    )
  )
  res <- contents_shinychat(result)
  expect_equal(res$request_call, 'test(x = 1, y = "test")')

  result@request@tool <- NULL
  res_no_tool <- contents_shinychat(result)
  expect_equal(
    jsonlite::fromJSON(res_no_tool$request_call),
    list(
      id = result@request@id,
      name = result@request@name,
      arguments = result@request@arguments
    )
  )
})

test_that("get_tool_result_display handles invalid formats", {
  # Test direct HTML warning
  result <- new_tool_result(
    extra = list(display = htmltools::tags$p("test"))
  )

  expect_snapshot(
    get_tool_result_display(result)
  )

  # Test non-list warning
  result <- new_tool_result(
    extra = list(display = "invalid")
  )
  expect_snapshot(
    get_tool_result_display(result)
  )
})

test_that("tool_result_display basic format", {
  local_shinychat_tool_display(opt = "basic")
  result <- new_tool_result(
    value = list(x = 1),
    extra = list(display = list(text = "ignored in basic mode"))
  )
  expect_equal(
    tool_result_display(result),
    list(
      value = jsonlite::toJSON(list(x = 1), auto_unbox = TRUE, pretty = 2),
      value_type = "code"
    )
  )
})

test_that("tool_result_display rich format", {
  local_shinychat_tool_display(opt = "rich")
  result <- new_tool_result(
    value = "test",
    extra = list(
      display = list(
        html = "<p>html</p>",
        markdown = "**md**",
        text = "text"
      )
    )
  )
  expect_equal(
    tool_result_display(result),
    list(value = "<p>html</p>", value_type = "html")
  )
})

test_that("processes a Turn object", {
  # Create a turn with multiple content items
  turn <- ellmer::AssistantTurn(
    contents = list(
      ellmer::ContentText("Hello"),
      new_tool_request(),
      ellmer::ContentText("World")
    )
  )

  # Process turn contents
  results <- contents_shinychat(turn)
  expect_length(results, 3)
  expect_equal(results[[1]], "Hello")
  expect_s3_class(results[[2]], "shinychat_tool_request")
  expect_equal(results[[3]], "World")
})

test_that("consolidates adjacent turn types in a Chat object", {
  withr::local_options(OPENAI_API_KEY = "boop")
  chat <- ellmer::chat_openai()

  chat$set_turns(
    list(
      ellmer::AssistantTurn(
        contents = list(ellmer::ContentText("Hello"))
      ),
      ellmer::AssistantTurn(
        contents = list(ellmer::ContentText("World"))
      )
    )
  )

  messages <- contents_shinychat(chat)
  expect_length(messages, 1)
  expect_equal(messages[[1]]$role, "assistant")
  expect_equal(messages[[1]]$content, "Hello\n\nWorld")
})

test_that("doesn't consolidate adjacent turns with different roles in a Chat object", {
  withr::local_options(OPENAI_API_KEY = "boop")
  chat <- ellmer::chat_openai()

  chat$set_turns(
    list(
      ellmer::UserTurn(
        contents = list(ellmer::ContentText("Question"))
      ),
      ellmer::AssistantTurn(
        contents = list(ellmer::ContentText("Answer"))
      )
    )
  )

  messages <- contents_shinychat(chat)
  expect_length(messages, 2) # Previous consolidated message + 2 new messages
  expect_equal(messages[[1]]$role, "user")
  expect_equal(messages[[2]]$role, "assistant")
})

test_that("drops requests and moves results to assistant turn role in a Chat object", {
  withr::local_options(OPENAI_API_KEY = "boop")
  chat <- ellmer::chat_openai()

  chat$set_turns(
    list(
      ellmer::AssistantTurn(
        contents = list(
          ellmer::ContentText("Hello"),
          new_tool_request()
        )
      ),
      ellmer::UserTurn(
        contents = list(
          new_tool_result(value = "success")
        )
      )
    )
  )

  messages <- contents_shinychat(chat)
  expect_length(messages, 1)
  expect_equal(messages[[1]]$role, "assistant")

  # Verify tool requests are filtered but results appear
  expect_false(
    some(messages[[1]]$content, inherits, "shinychat_tool_request")
  )
  expect_true(
    some(messages[[1]]$content, inherits, "shinychat_tool_result")
  )
})

test_that("throws when a result does not have a `request` property", {
  expect_snapshot(
    error = TRUE,
    contents_shinychat(new_tool_result(request = NULL))
  )
})

test_that("throws for invalid tool display option", {
  withr::local_options(shinychat.tool_display = "invalid")
  expect_snapshot(
    error = TRUE,
    opt_shinychat_tool_display()
  )
})

test_that("throws for invalid tool display ennvar", {
  withr::local_envvar(SHINYCHAT_TOOL_DISPLAY = "invalid")
  expect_snapshot(
    error = TRUE,
    opt_shinychat_tool_display()
  )
})

test_that("warns when `display` is not a list", {
  result <- new_tool_result(
    request = new_tool_request(),
    extra = list(display = htmltools::tags$p("test"))
  )
  expect_snapshot(
    as.tags(contents_shinychat(result))
  )
})

#' Create a UI element for a markdown stream.
#'
#' @description
#' Creates a UI element for a [markdown_stream()]. A markdown stream can be
#' useful for displaying generative AI responses (outside of a chat interface),
#' streaming logs, or other use cases where chunks of content are generated
#' over time.
#'
#' @param id A unique identifier for this markdown stream.
#' @param ... Extra HTML attributes to include on the chat element
#' @param content A string of content to display before any streaming occurs.
#'   When `content_type` is Markdown or HTML, it may also be UI element(s) such
#'   as input and output bindings.
#' @param content_type The content type. Default is `"markdown"` (specifically,
#'   CommonMark). Supported content types include:
#'       * `"markdown"`: markdown text, specifically CommonMark
#'       * `"html"`: for rendering HTML content.
#'       * `"text"`: for plain text.
#' @param auto_scroll Whether to automatically scroll to the bottom of a
#'   scrollable container when new content is added. Default is True.
#' @param width The width of the UI element.
#' @param height The height of the UI element.
#'
#' @return A shiny tag object.
#'
#' @export
#' @seealso [markdown_stream()]
#'
output_markdown_stream <- function(
  id,
  ...,
  content = "",
  content_type = "markdown",
  auto_scroll = TRUE,
  width = "min(680px, 100%)",
  height = "auto"
) {
  # `content` is most likely a string, so avoid overhead in that case
  # (it's also important that we *don't escape HTML* here).
  if (is.character(content)) {
    ui <- list(html = paste(content, collapse = "\n"))
  } else {
    ui <- with_current_theme({
      htmltools::renderTags(pre_process_ui(content))
    })
  }

  htmltools::tag(
    "shiny-markdown-stream",
    rlang::list2(
      id = id,
      style = css(
        width = width,
        height = height,
        margin = "0 auto"
      ),
      content = ui[["html"]],
      "content-type" = content_type,
      "auto-scroll" = if (auto_scroll) "" else NULL,
      ...,
      ui[["dependencies"]],
      shinychat_deps()
    )
  )
}

#' Stream markdown content
#'
#' @description
#' Streams markdown content into a [output_markdown_stream()] UI element.  A
#' markdown stream can be useful for displaying generative AI responses (outside
#' of a chat interface), streaming logs, or other use cases where chunks of
#' content are generated over time.
#'
#' @param id The ID of the markdown stream to stream content to.
#' @param content_stream A string generator (e.g., [coro::generator()] or
#' [coro::async_generator()]), a string promise (e.g., [promises::promise()]),
#' or a string promise generator.
#' @param operation The operation to perform on the markdown stream. The default,
#' `"replace"`, will replace the current content with the new content stream.
#' The other option, `"append"`, will append the new content stream to the
#' existing content.
#'
#' @param session The Shiny session object.
#'
#' @return NULL
#'
#' @export
#' @examplesIf interactive()
#'
#' library(shiny)
#' library(coro)
#' library(bslib)
#' library(shinychat)
#'
#' # Define a generator that yields a random response
#' # (imagine this is a more sophisticated AI generator)
#' random_response_generator <- async_generator(function() {
#'   responses <- c(
#'     "What does that suggest to you?",
#'     "I see.",
#'     "I'm not sure I understand you fully.",
#'     "What do you think?",
#'     "Can you elaborate on that?",
#'     "Interesting question! Let's examine thi... **See more**"
#'   )
#'
#'   await(async_sleep(1))
#'   for (chunk in strsplit(sample(responses, 1), "")[[1]]) {
#'     yield(chunk)
#'     await(async_sleep(0.02))
#'   }
#' })
#'
#' ui <- page_fillable(
#'   actionButton("generate", "Generate response"),
#'   output_markdown_stream("stream")
#' )
#'
#' server <- function(input, output, session) {
#'   observeEvent(input$generate, {
#'     markdown_stream("stream", random_response_generator())
#'   })
#' }
#'
#' shinyApp(ui, server)
markdown_stream <- function(
  id,
  content_stream,
  operation = c("replace", "append"),
  session = getDefaultReactiveDomain()
) {
  stream <- as_generator(content_stream)

  operation <- match.arg(operation)

  result <- markdown_stream_impl(id, stream, operation, session)
  result <- chat_update_bookmark(id, result, session = session)

  # Handle erroneous result...
  promises::catch(result, function(reason) {
    shiny::showNotification(
      sprintf(
        "Error in markdown_stream('%s'): %s",
        id,
        conditionMessage(reason)
      ),
      type = "error",
      duration = NULL,
      closeButton = TRUE
    )
  })
  # ...but also return it, so the caller can also handle it if they want. Note
  # that we're not returning the result of `promises::catch`; we want to return
  # a rejected promise (so the caller can see the error) that was already
  # handled (so there's no "unhandled promise error" warning if the caller
  # chooses not to do anything with it).
  result
}

markdown_stream_impl <- NULL
rlang::on_load(
  markdown_stream_impl <- coro::async(function(id, stream, operation, session) {
    send_stream_message <- function(...) {
      session$sendCustomMessage(
        "shinyMarkdownStreamMessage",
        rlang::list2(id = id, ...)
      )
    }

    if (operation == "replace") {
      send_stream_message(content = "", operation = "replace")
    }

    send_stream_message(isStreaming = TRUE)

    on.exit({
      send_stream_message(isStreaming = FALSE)
    })

    for (msg in stream) {
      if (promises::is.promising(msg)) {
        msg <- await(msg)
      }
      if (coro::is_exhausted(msg)) {
        break
      }

      if (is.character(msg)) {
        # content is most likely a string, so avoid overhead in that case
        ui <- list(html = msg, deps = "[]")
      } else {
        ui <- process_ui(pre_process_ui(msg), session)
      }

      send_stream_message(
        content = ui[["html"]],
        operation = "append",
        html_deps = ui[["deps"]]
      )
    }

    invisible(NULL)
  })
)

library(htmltools)

test_that("Chat component markup", {
  expect_snapshot(output_markdown_stream("stream"))

  expect_snapshot({
    output_markdown_stream("stream", content = "Foo\nBar")
  })

  expect_snapshot({
    render_tags(
      output_markdown_stream(
        "stream",
        content = div("Hello", htmlDependency("foo", "1.0.0", ""))
      )
    )
  })

  # TODO: it'd be nice to mock the messages that send_stream_message() sends
})

library(shiny)
library(bslib)
library(ellmer)
library(shinychat)
library(weathR)

get_weather_forecast <- tool(
  function(lat, lon) {
    point_tomorrow(lat, lon, short = FALSE)
  },
  name = "get_weather_forecast",
  description = "Get the weather forecast for a location.",
  arguments = list(
    lat = type_number("Latitude"),
    lon = type_number("Longitude")
  ),
  annotations = tool_annotations(
    title = "Weather Forecast",
    icon = bsicons::bs_icon("globe-americas")
  )
)

ui <- function(req) {
  page_fillable(
    chat_mod_ui("chat")
  )
}

server <- function(input, output, session) {
  client <- ellmer::chat("openai/gpt-4.1-nano")
  # client <- ellmer::chat_ollama(model = "mistral:v0.3")
  client$register_tool(get_weather_forecast)
  chat_mod_server("chat", client)
}

shinyApp(ui, server, enableBookmarking = "url")

library(shiny)
library(bslib)
library(shinychat)

ui <- page_fillable(
  chat_ui("chat", fill = TRUE)
)

server <- function(input, output, session) {
  # Append a user message from the server on init
  # (same code path as bookmark restoration via client_set_ui)
  observe({
    chat_append("chat", "A user message", role = "user")
    chat_append("chat", "An assistant reply")
  })
}

shinyApp(ui, server)

# This is a stripped-down port of the ui.Chat feature in py-shiny. The main
# things it's missing are server-side state management, i.e. the py-shiny
# version will keep the list of messages for you, and will handle the
# trimming of the message history to fit within the context window; these
# are left for the caller to handle in the R version.

#' Create a greeting for a chat UI
#'
#' @description
#' Creates a greeting object for use with [chat_ui()] or [chat_set_greeting()].
#' A greeting is displayed when the chat first loads and is dismissed when the
#' user sends their first message.
#'
#' @param content The greeting content. Can be:
#'   * A string, interpreted as markdown.
#'   * An [htmltools::HTML()] object, rendered as raw HTML.
#'   * An htmltools tag or [htmltools::tagList()], including Shiny
#'     inputs/outputs.
#'   * A generator or promise (only valid when used with `chat_set_greeting()`).
#' @param dismissible Whether the greeting is automatically dismissed when the
#'   user sends a message. Defaults to `TRUE`.
#'
#' @returns An S3 object of class `"chat_greeting"`.
#'
#' @section Patterns:
#'
#' **Non-dismissible greeting** (stays visible after the user sends a message):
#'
#' ```r
#' chat_greeting("Please read our [terms of service](https://example.com).", dismissible = FALSE)
#' ```
#'
#' **Greeting with suggestion cards** (clickable chips that fill the input):
#'
#' ```r
#' chat_greeting(paste(
#'   "## Welcome!\n\n",
#'   "Try one of these:\n\n",
#'   '<span class="suggestion">Summarize my data</span>\n',
#'   '<span class="suggestion">Create a plot</span>\n',
#'   '<span class="suggestion">Explain this code</span>'
#' ))
#' ```
#'
#' **Greeting with HTML tags** (Shiny inputs/outputs):
#'
#' ```r
#' chat_greeting(htmltools::tagList(
#'   htmltools::h2("Welcome!"),
#'   shiny::selectInput("model", "Choose a model:", c("gpt-4o", "claude-3"))
#' ))
#' ```
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#' library(shinychat)
#'
#' ui <- page_fillable(
#'   chat_ui(
#'     "chat",
#'     greeting = chat_greeting("## Welcome!\n\nHow can I help you today?")
#'   )
#' )
#'
#' server <- function(input, output, session) {
#'   observeEvent(input$chat_user_input, {
#'     response <- paste0("You said: ", input$chat_user_input)
#'     chat_append("chat", response)
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @export
chat_greeting <- function(
  content,
  dismissible = TRUE
) {
  if (is.character(content) && length(content) > 1) {
    content <- paste(content, collapse = "\n")
  }
  structure(
    list(
      content = content,
      dismissible = dismissible
    ),
    class = "chat_greeting"
  )
}

#' Create a chat UI element
#'
#' @description
#' Inserts a chat UI element into a Shiny UI, which includes a scrollable
#' section for displaying chat messages, and an input field for the user to
#' enter new messages.
#'
#' To respond to user input, listen for `input$ID_user_input` (for example, if
#' `id="my_chat"`, user input will be at `input$my_chat_user_input`), and use
#' [chat_append()] to append messages to the chat.
#'
#' @section Greeting:
#'
#' A greeting is an optional welcome message shown before any conversation
#' messages. It is automatically dismissed when the user sends their first
#' message (unless created with `dismissible = FALSE`).
#'
#' **Static greeting.** Pass a string or [chat_greeting()] to the `greeting`
#' parameter:
#'
#' ```r
#' chat_ui("chat", greeting = "## Hello!\n\nHow can I help you today?")
#' ```
#'
#' **Dynamic greeting from the server.** Leave `greeting` unset and use
#' [chat_set_greeting()] from your server function. This is useful when the
#' greeting depends on session state or is generated by a model.
#'
#' **`greeting_requested` input.** When the chat is visible on the page, has
#' no messages, and has no greeting set, Shiny fires
#' `input$<id>_greeting_requested` (e.g. `input$chat_greeting_requested` for
#' `chat_ui("chat")`). The value is an event counter suitable for
#' [shiny::observeEvent()]. Use it to trigger server-side greeting generation:
#'
#' ```r
#' observeEvent(input$chat_greeting_requested, {
#'   stream <- chat_client$stream_async("Generate a short welcome message.")
#'   chat_set_greeting("chat", chat_greeting(stream))
#' })
#' ```
#'
#' This input fires when the chat component is first viewed on the page and
#' empty, and again after [chat_clear()] `(greeting = TRUE)`, enabling a
#' regenerate pattern where clearing the greeting automatically triggers a
#' fresh one.
#'
#' @param id The ID of the chat element
#' @param ... Extra HTML attributes to include on the chat element
#' @param messages A list of messages to prepopulate the chat with. Each
#'   message can be one of the following:
#'
#'   * A string, which is interpreted as markdown and rendered to HTML on
#'     the client.
#'     * To prevent interpreting as markdown, mark the string as
#'       [htmltools::HTML()].
#'   * A UI element.
#'     * This includes [htmltools::tagList()], which take UI elements
#'       (including strings) as children. In this case, strings are still
#'       interpreted as markdown as long as they're not inside HTML.
#'   * A named list of `content` and `role`. The `content` can contain content
#'     as described above, and the `role` can be "assistant" or "user".
#'
#' @param greeting An optional greeting to display when the chat first loads.
#'   Can be a [chat_greeting()] object, or a plain string (which is
#'   auto-wrapped with default options). The greeting is dismissed when the
#'   user sends their first message. For example:
#'   `greeting = chat_greeting("## Hello!\n\nHow can I help you today?")`
#' @param placeholder The placeholder text for the chat's user input field
#' @param width The CSS width of the chat element
#' @param height The CSS height of the chat element
#' @param fill Whether the chat element should try to vertically fill its
#'   container, if the container is
#'   [fillable](https://rstudio.github.io/bslib/articles/filling/index.html)
#' @param icon_assistant The icon to use for the assistant chat messages.
#'   Can be HTML or a tag in the form of [htmltools::HTML()] or
#'   [htmltools::tags()]. If `None`, a default robot icon is used.
#' @param enable_cancel If `TRUE`, show a stop button during streaming that
#'   allows the user to cancel the in-progress response. When using
#'   [chat_mod_server()], cancellation is wired up automatically. For manual
#'   usage with `chat_ui()`, observe `input$<id>_cancel` to handle cancellation
#'   (e.g., by calling `ctrl$cancel()` on an ellmer `stream_controller()`).
#'   Defaults to `FALSE`.
#' @param footer Optional HTML content to display below the chat input.
#'   This can be any HTML content (tags, tag lists, or character strings).
#'   Useful for adding disclaimers, attribution, or other information.
#'   The footer text is styled slightly smaller and lighter than body text
#'   by default. Customize with CSS properties `--shiny-chat-footer-font-size`
#'   and `--shiny-chat-footer-color` on the chat container or footer element.
#' @section Thinking display:
#'
#' When a model produces reasoning or "thinking" tokens, shinychat renders them
#' in a collapsible panel above the response. The panel shows a live stream of
#' the model's reasoning while it thinks, then auto-collapses when the response
#' begins.
#'
#' Thinking display works automatically with any model that supports it. Two
#' paths are supported:
#'
#' 1. **ellmer's `ContentThinking` objects.** Models that provide a structured
#'    thinking API (e.g., Claude with extended thinking) emit `ContentThinking`
#'    objects when you stream with `stream = "content"`. shinychat detects these
#'    and routes them to the thinking panel. This is what `chat_append()` uses
#'    internally when you pass it an ellmer content stream.
#'
#' 2. **Raw `<thinking>` tags.** Many open-source and local models (DeepSeek,
#'    QwQ, Qwen, etc.) emit `<thinking>...</thinking>` tags directly in their
#'    markdown output. shinychat detects these tags during streaming and renders
#'    the enclosed text in the thinking panel with no extra configuration.
#'
#' ## Topic labels
#'
#' You can optionally get labeled sub-sections within the thinking panel by
#' asking the model to emit `<topic>...</topic>` tags in its reasoning. These
#' are extracted and rendered as section headings inside the thinking display,
#' and the current topic appears in the collapsed header as a live status.
#'
#' To use topic labels, add something like this to your system prompt:
#'
#' ```
#' When thinking through a problem, wrap brief topic labels in <topic> tags
#' to indicate what you're currently reasoning about. For example:
#' <topic>parsing the input</topic>
#' ```
#'
#' Topic labels are entirely optional. Without them, the thinking panel still
#' works -- it just won't have sub-section headings.
#'
#' @returns A Shiny tag object, suitable for inclusion in a Shiny UI
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#' library(shinychat)
#'
#' ui <- page_fillable(
#'   chat_ui("chat", fill = TRUE)
#' )
#'
#' server <- function(input, output, session) {
#'   observeEvent(input$chat_user_input, {
#'     # In a real app, this would call out to a chat client or API,
#'     # perhaps using the 'ellmer' package.
#'     response <- paste0(
#'       "You said:\n\n",
#'       "<blockquote>",
#'       htmltools::htmlEscape(input$chat_user_input),
#'       "</blockquote>"
#'     )
#'     chat_append("chat", response)
#'     chat_append("chat", stream)
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @export
chat_ui <- function(
  id,
  ...,
  messages = NULL,
  greeting = NULL,
  placeholder = "Enter a message...",
  width = "min(680px, 100%)",
  height = "auto",
  fill = TRUE,
  icon_assistant = NULL,
  enable_cancel = FALSE,
  footer = NULL
) {
  attrs <- rlang::list2(...)
  if (!all(nzchar(rlang::names2(attrs)))) {
    rlang::abort("All arguments in ... must be named.")
  }

  message_tags <- lapply(messages, function(x) {
    role <- "assistant"
    content <- x
    if (is.list(x) && ("content" %in% names(x))) {
      content <- x[["content"]]
      role <- x[["role"]] %||% role
    }

    # `content` is most likely a string, so avoid overhead in that case
    # (it's also important that we *don't escape HTML* here).
    if (is.character(content)) {
      ui <- list(html = paste(content, collapse = "\n"))
    } else {
      ui <- with_current_theme({
        htmltools::renderTags(pre_process_ui(content))
      })
    }

    tag(
      "shiny-chat-message",
      rlang::list2(
        `data-role` = role,
        content = ui[["html"]],
        icon = if (!is.null(icon_assistant)) as.character(icon_assistant),
        ui[["dependencies"]],
      )
    )
  })

  footer_tag <- NULL
  if (!is.null(footer)) {
    footer_tag <- tag("shiny-chat-footer", list(footer))
  }

  # Process greeting -------------------------------------------------------
  greeting_attr <- NULL
  greeting_deps <- list()

  if (!is.null(greeting)) {
    # Auto-wrap plain strings and HTML/tag content
    if (!inherits(greeting, "chat_greeting")) {
      greeting <- chat_greeting(greeting)
    }

    content <- greeting$content

    # Generators/promises are not allowed in chat_ui() (static context)
    if (
      inherits(content, "coro_generator_instance") ||
        promises::is.promising(content)
    ) {
      cli::cli_abort(
        c(
          "A generator or promise cannot be used as a static {.arg greeting} in {.fn chat_ui}.",
          i = "Use {.fn chat_set_greeting} from the server to set a dynamic or streaming greeting."
        )
      )
    }

    if (is.character(content) && !inherits(content, "html")) {
      # Plain markdown string
      greeting_content <- content
      greeting_content_type <- "markdown"
    } else if (
      inherits(content, "html") &&
        !inherits(content, c("shiny.tag", "shiny.tag.list"))
    ) {
      # htmltools::HTML() — raw HTML string
      greeting_content <- as.character(content)
      greeting_content_type <- "html"
    } else {
      # htmltools tag or tagList — render to HTML and collect deps
      rendered <- with_current_theme({
        htmltools::renderTags(pre_process_ui(content))
      })
      greeting_content <- rendered[["html"]]
      greeting_content_type <- "html"
      greeting_deps <- rendered[["dependencies"]]
    }

    greeting_payload <- list(
      content = greeting_content,
      content_type = greeting_content_type,
      options = list(
        dismissible = isTRUE(greeting$dismissible)
      )
    )
    greeting_attr <- jsonlite::toJSON(greeting_payload, auto_unbox = TRUE)
  }

  # Build UI ---------------------------------------------------------------
  res <- tag(
    "shiny-chat-container",
    rlang::list2(
      id = id,
      style = css(
        width = width,
        height = height
      ),
      placeholder = placeholder,
      fill = if (isTRUE(fill)) NA else NULL,
      `enable-cancel` = if (isTRUE(enable_cancel)) NA else NULL,
      # Also include icon on the parent so that when messages are dynamically added,
      # we know the default icon has changed
      `icon-assistant` = if (!is.null(icon_assistant)) {
        as.character(icon_assistant)
      },
      greeting = greeting_attr,
      ...,
      tag("shiny-chat-messages", message_tags),
      tag(
        "shiny-chat-input",
        list(id = paste0(id, "_user_input"), placeholder = placeholder)
      ),
      footer_tag,
      shinychat_deps(),
      htmltools::findDependencies(icon_assistant),
      greeting_deps
    )
  )

  if (isTRUE(fill)) {
    res <- bslib::as_fill_carrier(res)
  }

  tag_require(res, version = 5, caller = "chat_ui")
}

#' Append an assistant response (or user message) to a chat control
#'
#' @description
#' The `chat_append` function appends a message to an existing [chat_ui()]. The
#' `response` can be a string, string generator, string promise, or string
#' promise generator (as returned by the 'ellmer' package's `chat`, `stream`,
#' `chat_async`, and `stream_async` methods, respectively).
#'
#' This function should be called from a Shiny app's server. It is generally
#' used to append the client's response to the chat, while user messages are
#' added to the chat UI automatically by the front-end. You'd only need to use
#' `chat_append(role="user")` if you are programmatically generating queries
#' from the server and sending them on behalf of the user, and want them to be
#' reflected in the UI.
#'
#' # Error handling
#'
#' If the `response` argument is a generator, promise, or promise generator, and
#' an error occurs while producing the message (e.g., an iteration in
#' `stream_async` fails), the promise returned by `chat_append` will reject with
#' the error. If the `chat_append` call is the last expression in a Shiny
#' observer, shinychat will log the error message and show a message that the
#' error occurred in the chat UI.
#'
#' @param id The ID of the chat element
#' @param response The message or message stream to append to the chat element.
#'   The actual message content can one of the following:
#'
#'   * A string, which is interpreted as markdown and rendered to HTML on
#'     the client.
#'     * To prevent interpreting as markdown, mark the string as
#'       [htmltools::HTML()].
#'   * A UI element.
#'     * This includes [htmltools::tagList()], which take UI elements
#'       (including strings) as children. In this case, strings are still
#'       interpreted as markdown as long as they're not inside HTML.
#'
#' @param role The role of the message (either "assistant" or "user"). Defaults
#'   to "assistant".
#' @param icon An optional icon to display next to the message, currently only
#'   used for assistant messages. The icon can be any HTML element (e.g., an
#'   [htmltools::img()] tag) or a string of HTML.
#' @param session The Shiny session object
#'
#' @returns Returns a promise that resolves to the contents of the stream, or an
#'   error. This promise resolves when the message has been successfully sent to
#'   the client; note that it does not guarantee that the message was actually
#'   received or rendered by the client. The promise rejects if an error occurs
#'   while processing the response (see the "Error handling" section).
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(coro)
#' library(bslib)
#' library(shinychat)
#'
#' # Dumbest chatbot in the world: ignores user input and chooses
#' # a random, vague response.
#' fake_chatbot <- async_generator(function(input) {
#'   responses <- c(
#'     "What does that suggest to you?",
#'     "I see.",
#'     "I'm not sure I understand you fully.",
#'     "What do you think?",
#'     "Can you elaborate on that?",
#'     "Interesting question! Let's examine thi... **See more**"
#'   )
#'
#'   await(async_sleep(1))
#'   for (chunk in strsplit(sample(responses, 1), "")[[1]]) {
#'     yield(chunk)
#'     await(async_sleep(0.02))
#'   }
#' })
#'
#' ui <- page_fillable(
#'   chat_ui("chat", fill = TRUE)
#' )
#'
#' server <- function(input, output, session) {
#'   observeEvent(input$chat_user_input, {
#'     response <- fake_chatbot(input$chat_user_input)
#'     chat_append("chat", response)
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @export
chat_append <- function(
  id,
  response,
  role = c("assistant", "user"),
  icon = NULL,
  session = getDefaultReactiveDomain()
) {
  check_active_session(session)
  role <- match.arg(role)

  stream <- as_generator(response)
  chat_append_stream(id, stream, role = role, icon = icon, session = session)
}

#' Low-level function to append a message to a chat control
#'
#' For advanced users who want to control the message chunking behavior. Most
#' users should use [chat_append()] instead.
#'
#' @param id The ID of the chat element
#' @param msg The message to append. Should be a named list with `role` and
#'   `content` fields. The `role` field should be either "user" or "assistant".
#'   The `content` field should be a string containing the message content, in
#'   Markdown format.
#' @param chunk Whether `msg` is just a chunk of a message, and if so, what
#'   type. If `FALSE`, then `msg` is a complete message. If `"start"`, then
#'   `msg` is the first chunk of a multi-chunk message. If `"end"`, then `msg`
#'   is the last chunk of a multi-chunk message. If `TRUE`, then `msg` is an
#'   intermediate chunk of a multi-chunk message. Default is `FALSE`.
#' @param operation The operation to perform on the message. If `"append"`,
#'   then the new content is appended to the existing message content. If
#'   `"replace"`, then the existing message content is replaced by the new
#'   content. Ignored if `chunk` is `FALSE`.
#' @param icon An optional icon to display next to the message, currently only
#'   used for assistant messages. The icon can be any HTML element (e.g.,
#'   [htmltools::img()] tag) or a string of HTML.
#' @param session The Shiny session object
#'
#' @returns Returns nothing (\code{invisible(NULL)}).
#'
#' @importFrom shiny getDefaultReactiveDomain
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(coro)
#' library(bslib)
#' library(shinychat)
#'
#' # Dumbest chatbot in the world: ignores user input and chooses
#' # a random, vague response.
#' fake_chatbot <- async_generator(function(id, input) {
#'   responses <- c(
#'     "What does that suggest to you?",
#'     "I see.",
#'     "I'm not sure I understand you fully.",
#'     "What do you think?",
#'     "Can you elaborate on that?",
#'     "Interesting question! Let's examine thi... **See more**"
#'   )
#'
#'   # Use low-level chat_append_message() to temporarily set a progress message
#'   chat_append_message(id, list(role = "assistant", content = "_Thinking..._ "))
#'   await(async_sleep(1))
#'   # Clear the progress message
#'   chat_append_message(id, list(role = "assistant", content = ""), operation = "replace")
#'
#'   for (chunk in strsplit(sample(responses, 1), "")[[1]]) {
#'     yield(chunk)
#'     await(async_sleep(0.02))
#'   }
#' })
#'
#' ui <- page_fillable(
#'   chat_ui("chat", fill = TRUE)
#' )
#'
#' server <- function(input, output, session) {
#'   observeEvent(input$chat_user_input, {
#'     response <- fake_chatbot("chat", input$chat_user_input)
#'     chat_append("chat", response)
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @export
chat_append_message <- function(
  id,
  msg,
  chunk = TRUE,
  operation = c("append", "replace"),
  icon = NULL,
  session = getDefaultReactiveDomain()
) {
  check_active_session(session)

  if (!is.list(msg)) {
    rlang::abort("`msg` must be a named list with 'role' and 'content' fields")
  }
  if (!isTRUE(msg[["role"]] %in% c("user", "assistant"))) {
    warning("Invalid role argument; must be 'user' or 'assistant'")
    return(invisible(NULL))
  }

  if (!isFALSE(chunk)) {
    if (chunk == "start") {
      chunk_type <- "start"
    } else if (chunk == "end") {
      chunk_type <- "end"
    } else if (isTRUE(chunk)) {
      chunk_type <- "intermediate"
    } else {
      rlang::abort("Invalid chunk argument")
    }
  } else {
    chunk_type <- "complete"
  }

  content <- msg[["content"]]
  is_thinking <- inherits(content, "shinychat_thinking")
  if (is_thinking) {
    content <- unclass(content)
  }
  is_html <- inherits(
    content,
    c(
      "shiny.tag",
      "shiny.tag.list",
      "html",
      "htmlwidget",
      "shinychat_tool_card"
    )
  )
  content_type <- if (is_thinking) {
    "thinking"
  } else if (is_html) {
    "html"
  } else {
    "markdown"
  }

  operation <- match.arg(operation)

  if (is.character(content) && !is_html) {
    # content is most likely a string, so avoid overhead in that case
    ui <- list(html = content, deps = NULL)
  } else {
    # process_ui() does *not* render markdown->HTML, but it does:
    # 1. Extract and register HTMLdependency()s with the session.
    # 2. Returns a HTML string representation of the TagChild
    #    (i.e., `div()` -> `"<div>"`).
    ui <- process_ui(pre_process_ui(content), session)
  }

  msg_content <- ui[["html"]]
  if (is_html) {
    # Surround with blank lines so the markdown parser treats
    # block-level custom elements correctly.
    msg_content <- paste0("\n\n", msg_content, "\n\n")
  }

  html_deps <- ui[["deps"]]

  icon_str <- if (!is.null(icon)) as.character(icon) else NULL

  if (chunk_type == "start") {
    message_payload <- list(
      role = msg[["role"]],
      content = msg_content,
      content_type = content_type
    )
    if (!is.null(icon_str)) {
      message_payload$icon <- icon_str
    }
    action <- list(type = "chunk_start", message = message_payload)
    send_chat_action(
      id,
      action = action,
      html_deps = html_deps,
      session = session
    )
  } else if (chunk_type == "end") {
    if (nzchar(msg_content)) {
      chunk_action <- list(
        type = "chunk",
        content = msg_content,
        operation = operation,
        content_type = content_type
      )
      send_chat_action(
        id,
        action = chunk_action,
        html_deps = html_deps,
        session = session
      )
    }
    send_chat_action(id, action = list(type = "chunk_end"), session = session)
  } else if (chunk_type == "intermediate") {
    action <- list(
      type = "chunk",
      content = msg_content,
      operation = operation,
      content_type = content_type
    )
    send_chat_action(
      id,
      action = action,
      html_deps = html_deps,
      session = session
    )
  } else {
    # chunk_type == "complete"
    message_payload <- list(
      role = msg[["role"]],
      content = msg_content,
      content_type = content_type
    )
    if (!is.null(icon_str)) {
      message_payload$icon <- icon_str
    }
    action <- list(type = "message", message = message_payload)
    send_chat_action(
      id,
      action = action,
      html_deps = html_deps,
      session = session
    )
  }

  invisible(NULL)
}

chat_append_stream <- function(
  id,
  stream,
  role = "assistant",
  icon = NULL,
  session = getDefaultReactiveDomain()
) {
  result <- chat_append_stream_impl(id, stream, role, icon, session)
  result <- chat_update_bookmark(id, result, session = session)
  # Handle erroneous result...
  result <- promises::catch(result, function(reason) {
    # ...but rethrow the error as a silent error, so the caller can also handle
    # it if they want, but it won't bring down the app.
    class(reason) <- c("shiny.silent.error", class(reason))
    cnd_signal(reason)
  })

  promises::catch(result, function(reason) {
    chat_append_message(
      id,
      list(
        role = role,
        content = sanitized_chat_error(reason)
      ),
      chunk = "end",
      operation = "append",
      session = session
    )
    rlang::warn(
      sprintf(
        "ERROR: An error occurred in `chat_append_stream(id=\"%s\")`",
        session$ns(id)
      ),
      parent = reason
    )
  })

  # Note that we're not returning the result of `promises::catch()`, because we
  # want to return a rejected promise so the caller can see the error. But we
  # use the `catch()` both to make the error visible to the user *and* to ensure
  # there's no "unhandled promise error" warning if the caller chooses not to do
  # anything with it.
  result
}

utils:::globalVariables(c("generator_env", "exits", "yield"))

chat_append_stream_impl <- NULL
rlang::on_load(
  chat_append_stream_impl <- coro::async(function(
    id,
    stream,
    role = "assistant",
    icon = NULL,
    session = shiny::getDefaultReactiveDomain()
  ) {
    chat_append_ <- function(content, chunk = TRUE, ...) {
      chat_append_message(
        id,
        msg = list(role = role, content = content),
        operation = "append",
        chunk = chunk,
        session = session,
        ...
      )
    }

    chat_append_("", chunk = "start", icon = icon)

    res <- fastmap::fastqueue(200)

    for (msg in stream) {
      if (promises::is.promising(msg)) {
        msg <- await(msg)
      }
      if (coro::is_exhausted(msg)) {
        break
      }

      res$add(msg)

      if (S7::S7_inherits(msg, ellmer::ContentToolResult)) {
        if (!is.null(msg@request)) {
          send_chat_action(
            id,
            action = list(
              type = "hide_tool_request",
              requestId = msg@request@id
            ),
            session = session
          )
        }
      }

      if (S7::S7_inherits(msg, ellmer::Content)) {
        msg <- contents_shinychat(msg)
      }

      chat_append_(msg)
    }

    chat_append_("", chunk = "end")

    res <- res$as_list()
    if (every(res, is.character)) {
      paste(unlist(res), collapse = "")
    } else {
      res
    }
  })
)

#' Set the greeting for a chat UI
#'
#' @description
#' Sets or clears the greeting displayed in an existing [chat_ui()]. The
#' greeting is shown when the chat first loads and is dismissed when the user
#' sends their first message.
#'
#' Call `chat_set_greeting()` from the server to display a dynamic or streaming
#' greeting. This is typically used inside an [shiny::observeEvent()] watching
#' `input$<id>_greeting_requested` -- an event that fires when the chat is
#' visible, has no messages, and has no greeting set. See the **Greeting**
#' section in [chat_ui()] for details on `greeting_requested`.
#'
#' If the greeting has already been dismissed, calling this function updates
#' the greeting content but does not make it visible again. To show a new
#' greeting after dismissal, first clear the chat with
#' `chat_clear(id, greeting = TRUE)`.
#'
#' Pass `NULL` to clear the current greeting entirely.
#'
#' @param id The ID of the chat element
#' @param greeting The greeting to display. Can be:
#'   * `NULL` to clear the current greeting.
#'   * A plain string or [htmltools::HTML()] object (auto-wrapped with default
#'     options).
#'   * A [chat_greeting()] object.
#' @param session The Shiny session object
#'
#' @returns Returns `invisible(NULL)` for static and NULL greetings, or a
#'   promise for streaming greetings (resolves when streaming is complete).
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#' library(shinychat)
#'
#' # Static greeting set from the server
#' ui <- page_fillable(chat_ui("chat"))
#'
#' server <- function(input, output, session) {
#'   chat_set_greeting("chat", "## Welcome!\n\nHow can I help you today?")
#'
#'   observeEvent(input$chat_user_input, {
#'     chat_append("chat", paste0("You said: ", input$chat_user_input))
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#' library(shinychat)
#' library(coro)
#'
#' # Streaming greeting generated by a function
#' greeting_generator <- async_generator(function() {
#'   for (chunk in strsplit("## Hello!\n\nHow can I help you?", "")[[1]]) {
#'     yield(chunk)
#'     await(async_sleep(0.02))
#'   }
#' })
#'
#' ui <- page_fillable(chat_ui("chat"))
#'
#' server <- function(input, output, session) {
#'   chat_set_greeting("chat", chat_greeting(greeting_generator()))
#'
#'   observeEvent(input$chat_user_input, {
#'     chat_append("chat", paste0("You said: ", input$chat_user_input))
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#' library(shinychat)
#'
#' # LLM-generated greeting using greeting_requested
#' ui <- page_fillable(chat_ui("chat"))
#'
#' server <- function(input, output, session) {
#'   chat_client <- ellmer::chat_openai(model = "gpt-4o")
#'
#'   observeEvent(input$chat_greeting_requested, {
#'     stream <- chat_client$stream_async(
#'       "Generate a short, friendly welcome message."
#'     )
#'     chat_set_greeting("chat", chat_greeting(stream))
#'   })
#'
#'   observeEvent(input$chat_user_input, {
#'     stream <- chat_client$stream_async(input$chat_user_input)
#'     chat_append("chat", stream)
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#' library(shinychat)
#'
#' # Regenerate pattern: chat_clear(greeting = TRUE) triggers greeting_requested
#' ui <- page_fillable(
#'   chat_ui("chat"),
#'   actionButton("regenerate", "New greeting")
#' )
#'
#' server <- function(input, output, session) {
#'   observeEvent(input$chat_greeting_requested, {
#'     chat_set_greeting(
#'       "chat",
#'       paste("## Welcome!\n\nGenerated at", Sys.time())
#'     )
#'   })
#'
#'   observeEvent(input$regenerate, {
#'     chat_clear("chat", greeting = TRUE)
#'   })
#'
#'   observeEvent(input$chat_user_input, {
#'     chat_append("chat", paste0("You said: ", input$chat_user_input))
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @export
chat_set_greeting <- function(
  id,
  greeting,
  session = getDefaultReactiveDomain()
) {
  check_active_session(session)

  if (is.null(greeting)) {
    send_chat_action(
      id,
      action = list(type = "greeting_clear"),
      session = session
    )
    return(invisible(NULL))
  }

  if (!inherits(greeting, "chat_greeting")) {
    greeting <- chat_greeting(greeting)
  }

  content <- greeting$content
  options <- list(dismissible = isTRUE(greeting$dismissible))

  if (
    inherits(content, "coro_generator_instance") ||
      promises::is.promising(content)
  ) {
    stream <- as_generator(content)
    result <- chat_set_greeting_stream(id, stream, options, session)
    result <- promises::catch(result, function(reason) {
      class(reason) <- c("shiny.silent.error", class(reason))
      cnd_signal(reason)
    })
    promises::catch(result, function(reason) {
      send_chat_action(
        id,
        action = list(type = "greeting_end"),
        session = session
      )
      rlang::warn(
        sprintf(
          "ERROR: An error occurred in `chat_set_greeting(id=\"%s\")`",
          session$ns(id)
        ),
        parent = reason
      )
    })
    return(result)
  }

  if (is.function(content)) {
    cli::cli_abort(c(
      "{.fn chat_set_greeting} does not accept a function as greeting content.",
      "i" = "Pass the {.emph result} of calling your function, not the function itself.",
      "i" = "To use a greeting function with automatic lifecycle management, pass it to the {.arg greeting} argument of {.fn chat_mod_server}."
    ))
  }

  if (is.character(content) && !inherits(content, "html")) {
    greeting_content <- content
    greeting_content_type <- "markdown"
    html_deps <- NULL
  } else if (
    inherits(content, "html") &&
      !inherits(content, c("shiny.tag", "shiny.tag.list"))
  ) {
    greeting_content <- as.character(content)
    greeting_content_type <- "html"
    html_deps <- NULL
  } else {
    ui <- process_ui(pre_process_ui(content), session)
    greeting_content <- ui[["html"]]
    greeting_content_type <- "html"
    html_deps <- ui[["deps"]]
  }

  action <- list(
    type = "greeting",
    content = greeting_content,
    content_type = greeting_content_type,
    options = options
  )
  send_chat_action(
    id,
    action = action,
    html_deps = html_deps,
    session = session
  )
  invisible(NULL)
}

chat_set_greeting_stream <- NULL
rlang::on_load(
  chat_set_greeting_stream <- coro::async(function(
    id,
    stream,
    options,
    session = shiny::getDefaultReactiveDomain()
  ) {
    send_greeting_action <- function(action) {
      send_chat_action(id, action = action, session = session)
    }

    send_greeting_action(list(
      type = "greeting_start",
      content = "",
      content_type = "markdown",
      options = options
    ))

    for (msg in stream) {
      if (promises::is.promising(msg)) {
        msg <- await(msg)
      }
      if (coro::is_exhausted(msg)) {
        break
      }

      if (is.character(msg) && !inherits(msg, "html")) {
        ui <- list(html = msg, deps = NULL)
        chunk_content_type <- "markdown"
      } else {
        ui <- process_ui(pre_process_ui(msg), session)
        chunk_content_type <- "html"
      }

      send_chat_action(
        id,
        action = list(
          type = "greeting_chunk",
          content = ui[["html"]],
          operation = "append",
          content_type = chunk_content_type
        ),
        html_deps = ui[["deps"]],
        session = session
      )
    }

    send_greeting_action(list(type = "greeting_end"))
    invisible(NULL)
  })
)

#' Clear all messages from a chat control
#'
#' @description
#' Removes all messages from the chat UI. Set `greeting = TRUE` to also
#' clear the greeting, which re-triggers `greeting_requested` (see the
#' **Greeting** section in [chat_ui()]).
#'
#' @param id The ID of the chat element
#' @param greeting If `TRUE`, also clears the greeting. When the greeting is
#'   cleared, `greeting_requested` will fire again (if the chat is visible),
#'   allowing the server to generate a new greeting.
#' @param session The Shiny session object
#'
#' @export
#' @examplesIf interactive()
#'
#' library(shiny)
#' library(bslib)
#'
#' ui <- page_fillable(
#'   chat_ui("chat", fill = TRUE),
#'   actionButton("clear", "Clear chat")
#' )
#'
#' server <- function(input, output, session) {
#'   observeEvent(input$clear, {
#'     chat_clear("chat")
#'   })
#'
#'   observeEvent(input$chat_user_input, {
#'     response <- paste0("You said: ", input$chat_user_input)
#'     chat_append("chat", response)
#'   })
#' }
#'
#' shinyApp(ui, server)
#'
#' @examplesIf interactive()
#'
#' library(shiny)
#' library(bslib)
#' library(shinychat)
#'
#' # Regenerate greeting on clear
#' ui <- page_fillable(
#'   chat_ui("chat"),
#'   actionButton("new_chat", "New chat")
#' )
#'
#' server <- function(input, output, session) {
#'   observeEvent(input$chat_greeting_requested, {
#'     chat_set_greeting("chat", "## Welcome!\n\nHow can I help?")
#'   })
#'
#'   observeEvent(input$new_chat, {
#'     # Clearing with greeting = TRUE triggers greeting_requested again
#'     chat_clear("chat", greeting = TRUE)
#'   })
#'
#'   observeEvent(input$chat_user_input, {
#'     chat_append("chat", paste0("You said: ", input$chat_user_input))
#'   })
#' }
#'
#' shinyApp(ui, server)
chat_clear <- function(
  id,
  greeting = FALSE,
  session = getDefaultReactiveDomain()
) {
  check_active_session(session)
  action <- list(type = "clear")
  if (isTRUE(greeting)) {
    action$greeting <- TRUE
  }
  send_chat_action(id, action = action, session = session)
}

#' Update the user input of a chat control
#'
#' @param id The ID of the chat element
#' @param ... Currently unused, but reserved for future use.
#' @param value The value to set the user input to. If `NULL`, the input will not be updated.
#' @param placeholder The placeholder text for the user input
#' @param submit Whether to automatically submit the text for the user. Requires `value`.
#' @param focus Whether to move focus to the input element. Requires `value`.
#' @param session The Shiny session object
#'
#' @export
#' @examplesIf interactive()
#' library(shiny)
#' library(bslib)
#' library(shinychat)
#'
#' ui <- page_fillable(
#'   chat_ui("chat"),
#'   layout_columns(
#'     fill = FALSE,
#'     actionButton("update_placeholder", "Update placeholder"),
#'     actionButton("update_value", "Update user input")
#'   )
#' )
#'
#' server <- function(input, output, session) {
#'   observeEvent(input$update_placeholder, {
#'     update_chat_user_input("chat", placeholder = "New placeholder text")
#'   })
#'
#'   observeEvent(input$update_value, {
#'     update_chat_user_input("chat", value = "New user input", focus = TRUE)
#'   })
#'
#'   observeEvent(input$chat_user_input, {
#'     response <- paste0("You said: ", input$chat_user_input)
#'     chat_append("chat", response)
#'   })
#' }
#'
#' shinyApp(ui, server)

update_chat_user_input <- function(
  id,
  ...,
  value = NULL,
  placeholder = NULL,
  submit = FALSE,
  focus = FALSE,
  session = getDefaultReactiveDomain()
) {
  rlang::check_dots_empty()
  check_active_session(session)

  if (is.null(value) && (submit || focus)) {
    rlang::abort(
      "An input `value` must be provided when `submit` or `focus` are `TRUE`."
    )
  }

  vals <- drop_nulls(
    list(
      value = value,
      placeholder = placeholder,
      submit = submit,
      focus = focus
    )
  )

  action <- c(list(type = "update_input"), vals)

  send_chat_action(id, action = action, session = session)
}

library(htmltools)

# Helper: create a mock session and collect custom messages sent to it.
# Uses an environment so the message list is captured by reference.
mock_session_with_spy <- function() {
  sess <- shiny::MockShinySession$new()
  spy_env <- new.env(parent = emptyenv())
  spy_env$messages <- list()
  sess$sendCustomMessage <- function(type, msg) {
    spy_env$messages[[length(spy_env$messages) + 1]] <- list(
      type = type,
      message = msg
    )
  }
  list(session = sess, spy_env = spy_env)
}

# Return the custom messages recorded by a spy session.
spy_messages <- function(spy) spy$spy_env$messages


# ── chat_greeting() ────────────────────────────────────────────────────────────

test_that("chat_greeting() returns class 'chat_greeting' with all fields", {
  g <- chat_greeting("Hello")
  expect_s3_class(g, "chat_greeting")
  expect_equal(g$content, "Hello")
  expect_true(g$dismissible)
})

test_that("chat_greeting() stores non-default option values", {
  g <- chat_greeting("Hi", dismissible = FALSE)
  expect_false(g$dismissible)
})

test_that("chat_greeting() accepts HTML() content", {
  g <- chat_greeting(HTML("<b>bold</b>"))
  expect_s3_class(g, "chat_greeting")
  expect_s3_class(g$content, "html")
  expect_equal(as.character(g$content), "<b>bold</b>")
})

test_that("chat_greeting() accepts htmltools tag content", {
  tag_content <- tags$div("Welcome")
  g <- chat_greeting(tag_content)
  expect_s3_class(g, "chat_greeting")
  expect_s3_class(g$content, "shiny.tag")
})


# ── chat_ui(greeting = ...) ───────────────────────────────────────────────────

test_that("chat_ui() plain string greeting produces markdown content_type and dismissible", {
  ui <- chat_ui("chat", greeting = "## Hello")
  attr_raw <- ui$attribs$greeting
  expect_false(is.null(attr_raw))
  payload <- jsonlite::fromJSON(attr_raw)
  expect_equal(payload$content, "## Hello")
  expect_equal(payload$content_type, "markdown")
  expect_true(payload$options$dismissible)
})

test_that("chat_ui() chat_greeting with dismissible=FALSE serializes correctly", {
  g <- chat_greeting("## Hi", dismissible = FALSE)
  ui <- chat_ui("chat", greeting = g)
  payload <- jsonlite::fromJSON(ui$attribs$greeting)
  expect_equal(payload$content, "## Hi")
  expect_equal(payload$content_type, "markdown")
  expect_false(payload$options$dismissible)
})

test_that("chat_ui() HTML() greeting produces html content_type", {
  ui <- chat_ui("chat", greeting = HTML("<b>hi</b>"))
  payload <- jsonlite::fromJSON(ui$attribs$greeting)
  expect_equal(payload$content_type, "html")
  expect_equal(payload$content, "<b>hi</b>")
})

test_that("chat_ui() tag greeting produces html content_type and attaches dependencies", {
  tag_content <- tags$div("Hello", class = "greeting")
  ui <- chat_ui("chat", greeting = tag_content)
  payload <- jsonlite::fromJSON(ui$attribs$greeting)
  expect_equal(payload$content_type, "html")
  deps <- htmltools::findDependencies(ui)
  dep_names <- vapply(deps, `[[`, character(1), "name")
  expect_true("shinychat" %in% dep_names)
})

test_that("chat_ui() tag with explicit htmlDependency attaches that dep", {
  tag_content <- tags$div(
    "Hello",
    htmlDependency("my-dep", "1.0.0", ".")
  )
  ui <- chat_ui("chat", greeting = tag_content)
  deps <- htmltools::findDependencies(ui)
  dep_names <- vapply(deps, `[[`, character(1), "name")
  expect_true("my-dep" %in% dep_names)
})

test_that("chat_ui() rejects generator as greeting", {
  gen <- coro::async_generator(function() yield("x"))
  expect_error(
    chat_ui("chat", greeting = gen()),
    regexp = "generator or promise"
  )
})

test_that("chat_ui() snapshot for plain string greeting", {
  expect_snapshot(chat_ui("chat", greeting = "## Welcome!"))
})

test_that("chat_ui() snapshot for chat_greeting with dismissible=FALSE", {
  expect_snapshot(chat_ui(
    "chat",
    greeting = chat_greeting("## Hi", dismissible = FALSE)
  ))
})


# ── chat_set_greeting() ───────────────────────────────────────────────────────

test_that("chat_set_greeting() NULL sends greeting_clear action", {
  spy <- mock_session_with_spy()
  shiny::withReactiveDomain(spy$session, {
    chat_set_greeting("chat", NULL, session = spy$session)
  })
  msgs <- spy_messages(spy)
  expect_length(msgs, 1)
  expect_equal(msgs[[1]]$type, "shinyChatMessage")
  expect_equal(msgs[[1]]$message$action$type, "greeting_clear")
})

test_that("chat_set_greeting() plain string sends greeting action with markdown content_type", {
  spy <- mock_session_with_spy()
  shiny::withReactiveDomain(spy$session, {
    chat_set_greeting("chat", "Hello", session = spy$session)
  })
  msgs <- spy_messages(spy)
  expect_length(msgs, 1)
  action <- msgs[[1]]$message$action
  expect_equal(action$type, "greeting")
  expect_equal(action$content_type, "markdown")
  expect_true(action$options$dismissible)
})

test_that("chat_set_greeting() HTML() content sends html content_type", {
  spy <- mock_session_with_spy()
  shiny::withReactiveDomain(spy$session, {
    chat_set_greeting(
      "chat",
      chat_greeting(HTML("<b>hi</b>")),
      session = spy$session
    )
  })
  msgs <- spy_messages(spy)
  action <- msgs[[1]]$message$action
  expect_equal(action$type, "greeting")
  expect_equal(action$content_type, "html")
  expect_equal(action$content, "<b>hi</b>")
})

test_that("chat_set_greeting() generator sends greeting_start, greeting_chunk(s), greeting_end", {
  spy <- mock_session_with_spy()
  chunks <- c("He", "ll", "o")
  gen <- coro::async_generator(function() {
    for (ch in chunks) {
      yield(ch)
    }
  })
  shiny::withReactiveDomain(spy$session, {
    p <- chat_set_greeting("chat", chat_greeting(gen()), session = spy$session)
    done <- FALSE
    promises::then(
      p,
      function(x) {
        done <<- TRUE
      },
      function(e) {
        done <<- TRUE
      }
    )
    while (!done) {
      later::run_now(0.1)
    }
  })
  msgs <- spy_messages(spy)
  action_types <- vapply(msgs, function(m) m$message$action$type, character(1))
  expect_equal(action_types[[1]], "greeting_start")
  expect_equal(action_types[[length(action_types)]], "greeting_end")
  chunk_types <- action_types[action_types == "greeting_chunk"]
  expect_gte(length(chunk_types), 1)
  chunk_msgs <- Filter(
    function(m) m$message$action$type == "greeting_chunk",
    msgs
  )
  operations <- vapply(
    chunk_msgs,
    function(m) m$message$action$operation,
    character(1)
  )
  expect_true(all(operations == "append"))
})


test_that("chat_set_greeting() errors when given a function", {
  spy <- mock_session_with_spy()
  shiny::withReactiveDomain(spy$session, {
    expect_error(
      chat_set_greeting("chat", function() "hi", session = spy$session),
      "does not accept a function"
    )
  })
})


# ── chat_clear() greeting parameter ─────────────────────────────────────────

test_that("chat_clear() sends clear action without greeting field by default", {
  spy <- mock_session_with_spy()
  shiny::withReactiveDomain(spy$session, {
    chat_clear("chat", session = spy$session)
  })
  msgs <- spy_messages(spy)
  expect_length(msgs, 1)
  action <- msgs[[1]]$message$action
  expect_equal(action$type, "clear")
  expect_null(action$greeting)
})

test_that("chat_clear(greeting = TRUE) includes greeting in action", {
  spy <- mock_session_with_spy()
  shiny::withReactiveDomain(spy$session, {
    chat_clear("chat", greeting = TRUE, session = spy$session)
  })
  msgs <- spy_messages(spy)
  expect_length(msgs, 1)
  action <- msgs[[1]]$message$action
  expect_equal(action$type, "clear")
  expect_true(action$greeting)
})


# ── chat_mod_server() greeting function ──────────────────────────────────────

# Helper: minimal R6 mock that satisfies check_ellmer_chat() and chat_restore().
# Requires get_tools() so that the chat_restore set_ui observer does not crash
# and block subsequent observers.
mock_chat_client <- function(turns = list()) {
  R6::R6Class(
    "Chat",
    cloneable = TRUE,
    public = list(
      turns = turns,
      set_turns = function(t) {
        self$turns <- t
        invisible(self)
      },
      get_turns = function() self$turns,
      get_tools = function() list(),
      last_turn = function() NULL
    )
  )$new()
}

# Helper: suppress expected warnings from chat_restore()'s set_ui observer,
# which calls into ellmer internals not available on the mock client.
suppress_restore_warnings <- function(expr) {
  withCallingHandlers(
    expr,
    warning = function(w) {
      # Suppress expected warnings/errors from chat_restore's set_ui observer,
      # which calls ellmer internals not implemented on the mock client.
      if (
        grepl(
          "non-function|set_ui|tools|no applicable method",
          conditionMessage(w)
        )
      ) {
        invokeRestart("muffleWarning")
      }
    }
  )
}

test_that("named-arg detection: 'client' in formals identifies one-arg greeting", {
  expect_true("client" %in% names(formals(function(client) {})))
  expect_false("client" %in% names(formals(function() {})))
  expect_false("client" %in% names(formals(function(x) {})))
})

test_that("chat_mod_server() calls zero-arg greeting on chat_greeting_requested", {
  called <- FALSE
  suppress_restore_warnings(
    shiny::testServer(
      chat_mod_server,
      args = list(
        client = mock_chat_client(),
        greeting = function() {
          called <<- TRUE
          "## Hello"
        }
      ),
      {
        expect_false(called)
        session$setInputs(chat_greeting_requested = 1L)
        expect_true(called)
      }
    )
  )
})

test_that("chat_mod_server() zero-arg greeting is not called without input trigger", {
  called <- FALSE
  suppress_restore_warnings(
    shiny::testServer(
      chat_mod_server,
      args = list(
        client = mock_chat_client(),
        greeting = function() {
          called <<- TRUE
          "## Hello"
        }
      ),
      {
        # No setInputs — observer must not fire
        expect_false(called)
      }
    )
  )
})

test_that("chat_mod_server() calls one-arg greeting with a cloned client on chat_greeting_requested", {
  received_greeter <- NULL
  suppress_restore_warnings(
    shiny::testServer(
      chat_mod_server,
      args = list(
        client = mock_chat_client(),
        greeting = function(client) {
          received_greeter <<- client
          "## Hello"
        }
      ),
      {
        session$setInputs(chat_greeting_requested = 1L)
        expect_true(inherits(received_greeter, "Chat"))
      }
    )
  )
})

test_that("chat_mod_server() one-arg greeting receives a client with empty turns", {
  received_turns <- NULL
  client_with_turns <- mock_chat_client()
  client_with_turns$set_turns(list(list(
    role = "user",
    content = "prior message"
  )))
  suppress_restore_warnings(
    shiny::testServer(
      chat_mod_server,
      args = list(
        client = client_with_turns,
        greeting = function(client) {
          received_turns <<- client$get_turns()
          "## Hello"
        }
      ),
      {
        session$setInputs(chat_greeting_requested = 1L)
        expect_equal(length(received_turns), 0L)
      }
    )
  )
})

test_that("chat_mod_server() one-arg greeting does not clear original client turns", {
  client_with_turns <- mock_chat_client()
  client_with_turns$set_turns(list(list(
    role = "user",
    content = "prior message"
  )))
  suppress_restore_warnings(
    shiny::testServer(
      chat_mod_server,
      args = list(
        client = client_with_turns,
        greeting = function(client) "## Hello"
      ),
      {
        session$setInputs(chat_greeting_requested = 1L)
        expect_equal(length(client_with_turns$get_turns()), 1L)
      }
    )
  )
})

test_that("chat_mod_server() does not error with static string greeting", {
  expect_no_error(
    suppress_restore_warnings(
      shiny::testServer(
        chat_mod_server,
        args = list(client = mock_chat_client(), greeting = "## Static"),
        {}
      )
    )
  )
})

library(htmltools)

test_that("Chat component markup", {
  expect_snapshot(chat_ui("chat"))

  expect_snapshot({
    chat_ui("chat", messages = list("Foo", "Bar"))
  })

  expect_snapshot({
    chat_ui(
      "chat",
      messages = list(
        list(content = "Assistant", role = "assistant"),
        list(content = "User", role = "user")
      )
    )
  })

  expect_snapshot({
    chat_ui(
      "chat",
      messages = list(
        div("Hello"),
        span("world")
      )
    )
  })

  expect_snapshot({
    render_tags(
      chat_ui(
        "chat",
        messages = list(
          div("Hello", htmlDependency("foo", "1.0.0", "")),
          span("world")
        )
      )
    )
  })

  # Initial messages with react elements are island-split
  expect_snapshot({
    react_tag <- tags$div("react", `data-shinychat-react` = NA)
    chat_ui(
      "chat",
      messages = list(
        tagList(tags$div("before"), react_tag, tags$div("after"))
      )
    )
  })

  # TODO: it'd be nice to mock the shinyChatMessage custom messages
})

test_that("chat_append_stream() returns the stream contents as string if all text", {
  local_mocked_bindings(
    chat_append_message = coro::async(function(...) invisible())
  )

  stream <- coro::async_generator(function() {
    for (i in c("Hello", ",", " world", "!")) {
      yield(i)
    }
  })

  p <- chat_append_stream("chat", stream())
  res <- sync(p)

  expect_promise(p, "fulfilled")
  expect_equal(res, "Hello, world!")
})

test_that("chat_append_stream() returns the stream contents as list if not all text", {
  local_mocked_bindings(
    chat_append_message = coro::async(function(...) invisible())
  )

  stream <- coro::async_generator(function() {
    for (i in c("Hello", ",", " world", "!")) {
      yield(ellmer::ContentText(i))
    }
  })

  p <- chat_append_stream("chat", stream())
  res <- sync(p)

  expect_promise(p, "fulfilled")

  expect_true(is.list(res))
  expect_true(every(res, inherits, "ellmer::ContentText"))
  expect_equal(
    paste(map_chr(res, ellmer::contents_text), collapse = ""),
    "Hello, world!"
  )
})

test_that("chat_append_stream() handles errors in the stream", {
  local_mocked_bindings(
    chat_append_message = coro::async(function(...) invisible())
  )

  shiny::withReactiveDomain(shiny::MockShinySession$new(), {
    stream <- coro::async_generator(function() {
      for (i in 1:3) {
        yield(i)
      }
      stop("boom")
    })

    p <- chat_append_stream("chat", stream())
    expect_warning(
      res <- tryCatch(sync(p), error = identity),
      regexp = 'chat_append_stream'
    )

    expect_promise(p, "rejected")

    expect_s3_class(res, class = c("condition", "error"))
    expect_s3_class(res, class = "shiny.silent.error")
    expect_equal(conditionMessage(res), "boom")
  })
})

opt_shinychat_tool_display <- function() {
  choices <- c("none", "basic", "rich")

  opt <- getOption("shinychat.tool_display", default = NULL)
  if (!is.null(opt)) {
    opt <- arg_match(opt, choices, error_arg = "shinychat.tool_display")
    return(opt)
  }

  env <- Sys.getenv("SHINYCHAT_TOOL_DISPLAY", unset = "rich")
  arg_match(env, choices, error_arg = "SHINYCHAT_TOOL_DISPLAY")
}

#' Format ellmer content for shinychat
#'
#' @section Extending `contents_shinychat()`:
#'
#' You can extend `contents_shinychat()` to handle custom content types in your
#' application. `contents_shinychat()` is [an S7 generic][S7::new_generic]. If
#' you haven't worked with S7 before, you can learn more about S7 classes,
#' generics and methods in the [S7
#' documentation](https://rconsortium.github.io/S7/articles/S7.html).
#'
#' We'll work through a short example creating a custom display for the results
#' of a tool that gets local weather forecasts. We first need to create a custom
#' class that extends [ellmer::ContentToolResult].
#'
#' ```r
#' library(ellmer)
#'
#' WeatherToolResult <- S7::new_class(
#'   "WeatherToolResult",
#'   parent = ContentToolResult,
#'   properties = list(
#'     location_name = S7::class_character
#'   )
#' )
#' ```
#'
#' Next, we'll create a simple [ellmer::tool()] that gets the weather forecast
#' for a location and returns our custom `WeatherToolResult` class. The custom
#' class works just like a regular `ContentToolResult`, but it has an additional
#' `location_name` property.
#'
#' ```r
#' get_weather_forecast <- tool(
#'   function(lat, lon, location_name) {
#'     WeatherToolResult(
#'       weathR::point_tomorrow(lat, lon, short = FALSE),
#'       location_name = location_name
#'     )
#'   },
#'   name = "get_weather_forecast",
#'   description = "Get the weather forecast for a location.",
#'   arguments = list(
#'     lat = type_number("Latitude"),
#'     lon = type_number("Longitude"),
#'     location_name = type_string("Name of the location for display to the user")
#'   )
#' )
#' ```
#'
#' Finally, we can extend `contents_shinychat()` to render our custom content
#' class for display in the chat interface. The basic process is to define a
#' `contents_shinychat()` external generic and then implement a method for your
#' custom class.
#'
#' ```r
#' contents_shinychat <- S7::new_external_generic(
#'   package = "shinychat",
#'   name = "contents_shinychat",
#'   dispatch_args = "contents"
#' )
#'
#' S7::method(contents_shinychat, WeatherToolResult) <- function(content) {
#'   # Your custom rendering logic here
#' }
#' ```
#'
#' You can use this pattern to completely customize how the content is displayed
#' inside shinychat by returning HTML objects directly from this method.
#'
#' You can also use this pattern to build upon the default shinychat display for
#' tool requests and results. By using [S7::super()], you can create the
#' object shinychat uses for tool results (or tool requests), and then modify it
#' to suit your needs.
#'
#' ```r
#' S7::method(contents_shinychat, WeatherToolResult) <- function(content) {
#'   # Call the super method for ContentToolResult to get shinychat's defaults
#'   res <- contents_shinychat(S7::super(content, ContentToolResult))
#'
#'   # Then update the result object with more specific content
#'   # In this case, we render the tool result dataframe as a {gt} table...
#'   res$value <- gt::as_raw_html(gt::gt(content@value))
#'   res$value_type <- "html"
#'   # ...and update the tool result title to include the location name
#'   res$title <- paste("Weather Forecast for", content@location_name)
#'
#'   res
#' }
#' ```
#'
#' Note that you do **not** need to create a new class or extend
#' `contents_shinychat()` to customize the tool display. Rather, you can use the
#' strategies discussed in the [Tool Calling UI
#' article](https://posit-dev.github.io/shinychat/r/articles/tool-ui.html) to
#' customize the tool request and result display by providing a `display` list
#' in the `extra` argument of the tool result.
#'
#' @param content An [`ellmer::Content`] object.
#'
#' @return Returns text, HTML, or web component tags formatted for use in
#'   `chat_ui()`.
#'
#' @export
contents_shinychat <- S7::new_generic(
  "contents_shinychat",
  "content",
  function(content) {
    S7::S7_dispatch()
  }
)

S7::method(contents_shinychat, ellmer::Content) <- function(content) {
  # Fall back to html or markdown
  html <- ellmer::contents_html(content)
  if (!is.null(html)) {
    shiny::HTML(html)
  } else {
    ellmer::contents_markdown(content)
  }
}

S7::method(contents_shinychat, ellmer::ContentText) <- function(content) {
  content@text
}

S7::method(contents_shinychat, ellmer::ContentThinking) <- function(content) {
  structure(content@thinking, class = "shinychat_thinking")
}

new_tool_card <- function(type, request_id, tool_name, ...) {
  type <- arg_match(type, c("request", "result"))

  classes <- c(
    paste0("shinychat_tool_", type),
    "shinychat_tool_card"
  )

  dots <- dots_list(
    type = type,
    request_id = request_id,
    tool_name = tool_name,
    ...
  )

  structure(dots, class = classes)
}

#' @export
as.tags.shinychat_tool_card <- function(x, ...) {
  tag_name <- switch(
    x$type,
    request = "shiny-tool-request",
    result = "shiny-tool-result",
    cli::cli_abort(
      "shinychat tool card must have type {.val request} or {.val result}, not {.val {x$type}}."
    )
  )

  if (identical(x$value_type, "html") && !is.character(x$value)) {
    x$value <- as.tags(x$value)
  }
  if (!is.null(x$icon) && !is.character(x$icon)) {
    x$icon <- as.tags(x$icon)
  }
  if (!is.null(x$footer) && !is.character(x$footer)) {
    x$footer <- as.tags(x$footer)
  }

  names(x) <- gsub("_", "-", names(x))

  deps <- list(
    htmltools::findDependencies(x$value),
    htmltools::findDependencies(x$icon),
    htmltools::findDependencies(x$footer),
    shinychat_deps()
  )

  tag <- htmltools::tag(
    tag_name,
    dots_list(type = NULL, !!!x, !!!deps, .homonyms = "first")
  )
  htmltools::tagAppendAttributes(tag, `data-shinychat-react` = NA)
}

#' @export
format.shinychat_tool_card <- function(x, ...) {
  format(as.tags(x), ...)
}

#' @export
print.shinychat_tool_card <- function(x, ...) {
  tags <- as.tags(x)
  class(tags) <- c("bslib_fragment", class(tags))
  attr(tags, "bslib_page") <- function(...) {
    bslib::page_fluid(
      htmltools::div(
        class = "m-3",
        ...
      )
    )
  }
  print(tags, ...)
  invisible(x)
}

#' @exportS3Method knitr::knit_print
knit_print.shinychat_tool_card <- function(x, ...) {
  knitr::knit_print(as.tags(x))
}

S7::method(contents_shinychat, ellmer::ContentToolRequest) <- function(
  content
) {
  if (opt_shinychat_tool_display() == "none") {
    return(NULL)
  }

  tool <- content@tool

  new_tool_card(
    "request",
    request_id = content@id,
    tool_name = content@name,
    arguments = jsonlite::toJSON(content@arguments, auto_unbox = TRUE),
    intent = content@arguments[["_intent"]],
    tool_title = if (!is.null(tool)) tool@annotations$title
  )
}

S7::method(contents_shinychat, ellmer::ContentToolResult) <- function(content) {
  if (opt_shinychat_tool_display() == "none") {
    return(NULL)
  }

  if (is.null(content@request)) {
    cli::cli_abort(
      "`ContentToolResult` objects must have an associated `@request` property."
    )
  }

  display <- get_tool_result_display(content)
  annotations <- list()

  if (!is.null(content@request@tool)) {
    annotations <- content@request@tool@annotations
    request_call <- format(content@request, show = "call")
  } else {
    # formatting the request fails if tool is not present
    # (ellmer v0.3.0, tidyverse/ellmer#691)
    request_call <- jsonlite::toJSON(
      list(
        id = content@request@id,
        name = content@request@name,
        arguments = content@request@arguments
      ),
      auto_unbox = TRUE,
      pretty = 2
    )
  }

  new_tool_card(
    "result",
    request_id = content@request@id,
    request_call = request_call,
    status = if (tool_errored(content)) "error" else "success",
    tool_name = content@request@name,
    tool_title = display$title %||% annotations$title,
    icon = display$icon %||% annotations$icon,
    intent = content@request@arguments[["_intent"]],
    show_request = if (!isFALSE(display$show_request)) NA,
    expanded = if (isTRUE(display$open)) NA,
    full_screen = if (isTRUE(display$full_screen)) NA,
    footer = display$footer,
    !!!tool_result_display(content, display)
  )
}

get_tool_result_display <- function(content) {
  display <- content@extra$display
  request <- content@request

  if (is.null(display) || opt_shinychat_tool_display() == "basic") {
    return(list())
  }

  invalid_display_fmt <- "Invalid {.code @extra$display} format for {.code ContentToolResult} from {.fn {request@name}} (call id: {request@id})."

  if (
    inherits(display, c("html", "shiny.tag", "shiny.tag.list", "htmlwidgets"))
  ) {
    cli::cli_warn(
      c(
        invalid_display_fmt,
        "i" = "To display HTML content for tool results in {.pkg shinychat}, create a tool result with {.code extra = list(display = list(html = ...))}.",
        "i" = "You can also use {.code markdown} or {.code text} items in {.code display} to show Markdown or plain text, respectively."
      )
    )
    return(list())
  }

  # fmt: skip
  expected_fields <- c(
    "html",
    "markdown",
    "text",
    "show_request",
    "open",
    "full_screen",
    "title",
    "icon",
    "footer"
  )

  if (!is.list(display)) {
    cli::cli_warn(
      c(
        invalid_display_fmt,
        "x" = "Expected a list with fields {.or {.var {expected_fields}}}, not {.obj_type_friendly {display}}."
      )
    )
    return(list())
  }

  display
}

tool_result_display <- function(content, display = NULL) {
  display <- display %||% content@extra$display

  has_display <- !is.null(display) && is.list(display) && length(display) > 0
  use_basic_display <- opt_shinychat_tool_display() == "basic"

  if (tool_errored(content) || use_basic_display || !has_display) {
    return(tool_default_display(content))
  }

  if (is.list(display)) {
    has_type <- intersect(c("html", "markdown", "text"), names(display))
    if (length(has_type) > 0) {
      value_type <- has_type[1]
      return(list(value = display[[value_type]], value_type = value_type))
    }
  }

  tool_default_display(content)
}

# Copied from
# https://github.com/tidyverse/ellmer/blob/11cf1696/R/content.R#L292-L308
tool_errored <- function(x) !is.null(x@error)
tool_error_string <- function(x) {
  if (inherits(x@error, "condition")) conditionMessage(x@error) else x@error
}
tool_string_value <- function(x) {
  if (inherits(x@value, "AsIs")) {
    x@value
  } else if (inherits(x@value, "json")) {
    x@value
  } else if (is.character(x@value)) {
    paste(x@value, collapse = "\n")
  } else {
    jsonlite::toJSON(x@value, auto_unbox = TRUE, pretty = 2, force = TRUE)
  }
}

is_content_extra <- function(x) {
  is_content_image(x) || S7::S7_inherits(x, ellmer::ContentPDF)
}

is_content_image <- function(x) {
  S7::S7_inherits(x, ellmer::ContentImage)
}

as_content_extra_item <- function(x) {
  if (S7::S7_inherits(x, ellmer::ContentImageRemote)) {
    list(type = "image", src = x@url)
  } else if (S7::S7_inherits(x, ellmer::ContentImageInline)) {
    list(type = "image", src = paste0("data:", x@type, ";base64,", x@data))
  } else if (S7::S7_inherits(x, ellmer::ContentPDF)) {
    list(type = "pdf", filename = x@filename %||% "document.pdf")
  }
}

is_content <- function(x) {
  S7::S7_inherits(x, ellmer::Content)
}

as_content_extra_item_or_text <- function(x) {
  if (is_content_extra(x)) {
    as_content_extra_item(x)
  } else if (S7::S7_inherits(x, ellmer::ContentText)) {
    list(type = "text", value = x@text, value_type = "markdown")
  } else {
    list(type = "text", value = as.character(x), value_type = "markdown")
  }
}

tool_default_display <- function(content) {
  value <- content@value

  if (tool_errored(content)) {
    return(list(
      value = strip_ansi(tool_error_string(content)),
      value_type = "code"
    ))
  }

  if (is_content_extra(value)) {
    return(list(
      value = jsonlite::toJSON(
        list(as_content_extra_item(value)),
        auto_unbox = TRUE
      ),
      value_type = "content_extra"
    ))
  }

  if (is.list(value) && some(value, is_content)) {
    items <- map(value, as_content_extra_item_or_text)
    return(list(
      value = jsonlite::toJSON(items, auto_unbox = TRUE),
      value_type = "content_extra"
    ))
  }

  list(value = tool_string_value(content), value_type = "code")
}

S7::method(contents_shinychat, ellmer::Turn) <- function(content) {
  # Process all contents in the turn, filtering out empty results
  compact(map(content@contents, contents_shinychat))
}

S7::method(contents_shinychat, S7::new_S3_class(c("Chat", "R6"))) <- function(
  content
) {
  tools <- content$get_tools()

  # Process turns with tool request/result consolidation
  turns <- map(content$get_turns(), function(turn) {
    turn@contents <- map(turn@contents, function(x) {
      if (!S7::S7_inherits(x, ellmer::ContentToolResult)) {
        return(x)
      }
      if (!is.null(x@request@tool)) {
        return(x)
      }
      if (x@request@name %in% names(tools)) {
        x@request@tool <- tools[[x@request@name]]
      }
      x
    })

    # Turns containing only tool results are converted into assistant turns
    if (every(turn@contents, S7::S7_inherits, ellmer::ContentToolResult)) {
      if (packageVersion("ellmer") >= "0.3.2.9000") {
        turn <- ellmer::AssistantTurn(turn@contents)
      } else {
        turn@role <- "assistant"
      }
      return(turn)
    }

    # Filter out tool requests as they'll be shown in results
    is_tool_request <- map_lgl(
      turn@contents,
      S7::S7_inherits,
      ellmer::ContentToolRequest
    )
    turn@contents <- turn@contents[!is_tool_request]

    turn
  })

  # Consolidate adjacent turns with the same role
  turns <- reduce(
    turns,
    .init = list(),
    function(acc, turn) {
      if (length(acc) == 0) {
        return(list(turn))
      }

      last_turn <- acc[[length(acc)]]
      if (identical(last_turn@role, turn@role)) {
        acc[[length(acc)]]@contents <- c(last_turn@contents, turn@contents)
        return(acc)
      }

      c(acc, list(turn))
    }
  )

  # Convert turns to messages
  messages <- map(turns, function(turn) {
    content <- compact(contents_shinychat(turn))
    if (is.null(content) || identical(content, "")) {
      return(NULL)
    }
    if (every(content, is.character)) {
      content <- paste(unlist(content), collapse = "\n\n")
    }
    list(role = turn@role, content = content)
  })

  compact(messages)
}

library(shiny)
library(bslib)
library(shinychat)
library(fontawesome)

# Add resource path for images
addResourcePath("img", "img")

ui <- page_fillable(
  title = "Chat Icons",

  layout_columns(
    # Default Bot ----
    div(
      h2("Default Bot"),
      chat_ui(
        id = "chat_default",
        messages = list(
          list(
            content = "Hello! I'm Default Bot. How can I help you today?",
            role = "assistant"
          )
        ),
        icon_assistant = NULL # Uses default robot icon
      )
    ),

    # Animal Bot ----
    div(
      h2("Animal Bot"),
      chat_ui(
        id = "chat_animal",
        messages = list("Hello! I'm Animal Bot. How can I help you today?"),
        icon_assistant = fontawesome::fa("otter", title = "icon-otter")
      ),
      selectInput(
        "animal",
        "Animal",
        choices = c("Otter", "Hippo", "Frog", "Dove"),
        selected = "Otter"
      )
    ),

    # SVG Bot ----
    div(
      h2("SVG Bot"),
      chat_ui(
        id = "chat_svg",
        messages = list("Hello! I'm SVG Bot. How can I help you today?"),
        icon_assistant = HTML(
          '
          <svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-info-circle-fill icon-svg" viewBox="0 0 16 16">
            <path d="M8 16A8 8 0 1 0 8 0a8 8 0 0 0 0 16m.93-9.412-1 4.705c-.07.34.029.533.304.533.194 0 .487-.07.686-.246l-.088.416c-.287.346-.92.598-1.465.598-.703 0-1.002-.422-.808-1.319l.738-3.468c.064-.293.006-.399-.287-.47l-.451-.081.082-.381 2.29-.287zM8 5.5a1 1 0 1 1 0-2 1 1 0 0 1 0 2"/>
          </svg>
        '
        )
      )
    ),

    # Image Bot ----
    div(
      h2("Image Bot"),
      chat_ui(
        id = "chat_image",
        messages = list("Hello! I'm Image Bot. How can I help you today?"),
        icon_assistant = img(
          src = "img/grace-hopper.jpg",
          class = "icon-image grace-hopper"
        )
      ),
      selectInput(
        "image",
        "Image",
        choices = c("Grace Hopper", "Shiny"),
        selected = "Grace Hopper"
      )
    )
  )
)

server <- function(input, output, session) {
  # Default Bot ----
  observeEvent(input$chat_default_user_input, {
    req(input$chat_default_user_input)

    # Simulate delay
    Sys.sleep(1)

    chat_append(
      "chat_default",
      paste0("You said: ", input$chat_default_user_input)
    )
  })

  # Animal Bot ----
  observeEvent(input$chat_animal_user_input, {
    req(input$chat_animal_user_input)

    # Simulate delay
    Sys.sleep(1)

    animal <- tolower(input$animal)

    # Create icon based on selection
    if (animal == "otter") {
      # Use default icon (NULL)
      icon <- NULL
    } else {
      icon_map <- list(
        "hippo" = "hippo",
        "frog" = "frog",
        "dove" = "dove"
      )

      if (animal %in% names(icon_map)) {
        icon <- fontawesome::fa(
          icon_map[[animal]],
          # fontawesome doesn't support `class` argument, so we use `title`
          title = paste0("icon-", animal)
        )
      } else {
        icon <- NULL
      }
    }

    chat_append(
      "chat_animal",
      paste0(animal, " said: ", input$chat_animal_user_input),
      icon = icon
    )
  })

  # SVG Bot ----
  observeEvent(input$chat_svg_user_input, {
    req(input$chat_svg_user_input)

    chat_append(
      "chat_svg",
      paste0("You said: ", input$chat_svg_user_input)
    )
  })

  # Image Bot ----
  observeEvent(input$chat_image_user_input, {
    req(input$chat_image_user_input)

    # Create icon based on selection
    icon <- NULL
    if (input$image == "Shiny") {
      icon <- img(
        src = "img/shiny.png",
        class = "icon-image shiny"
      )
    }

    chat_append(
      "chat_image",
      paste0("You said: ", input$chat_image_user_input),
      icon = icon
    )
  })
}

shinyApp(ui, server)

library(shiny)
library(bslib)
library(ellmer)
library(leaflet)
library(shinychat)

tool_show_map <- tool(
  function(lat, lon, title, description) {
    popup_content <- sprintf(
      "<strong>%s</strong><br>%s",
      title,
      description
    )

    map <- leaflet() |>
      addTiles() |>
      addMarkers(lat = lat, lng = lon, popup = popup_content) |>
      setView(lng = lon, lat = lat, zoom = 13)

    ContentToolResult(
      value = "Map shown to the user.",
      extra = list(
        display = list(
          title = HTML(sprintf("Map of <b>%s</b>", title)),
          html = map,
          show_request = FALSE,
          open = TRUE,
          full_screen = TRUE
        )
      )
    )
  },
  name = "tool_show_map",
  description = r"(Show a map with a marker.

Use this tool whenever you're talking about a location with the user.
)",
  arguments = list(
    lat = type_number("Latitude"),
    lon = type_number("Longitude"),
    title = type_string("Title for the location"),
    description = type_string(
      "A short description of the location, based on the context of the conversation so far or your knowledge of the location."
    )
  ),
  annotations = list(
    title = "Map",
    icon = r"(<svg xmlns="http://www.w3.org/2000/svg" width="16" height="16" fill="currentColor" class="bi bi-geo-alt-fill" viewBox="0 0 16 16">
  <path d="M8 16s6-5.686 6-10A6 6 0 0 0 2 6c0 4.314 6 10 6 10m0-7a3 3 0 1 1 0-6 3 3 0 0 1 0 6"/>
</svg>)"
  )
)

ui <- function(req) {
  page_fillable(
    chat_mod_ui("chat")
  )
}

server <- function(input, output, session) {
  client <- ellmer::chat(
    "openai/gpt-4.1-nano",
    system_prompt = r"(
You're a helpful guide who can tell users about places and show them maps.

Anytime you mention a location, use the `tool_show_map` tool to show a map with a marker at the location. Don't make the user ask to see the map, just show it automatically when it'd be relevant to have a visual.)"
  )
  client$register_tool(tool_show_map)
  chat_mod_server("chat", client)
}

shinyApp(ui, server, enableBookmarking = "url")

# Given a promise-yielding expression, loop until it resolves or rejects.
# DON'T USE THIS TECHNIQUE IN SHINY, PLUMBER, OR HTTPUV CONTEXTS.
sync <- function(expr) {
  p <- force(expr)

  done <- FALSE
  success <- NULL
  error <- NULL

  promises::then(
    p,
    function(result) {
      success <<- result
      done <<- TRUE
    },
    function(err) {
      error <<- err
      done <<- TRUE
    }
  )

  while (!done) {
    later::run_now(0.25)
  }
  if (!is.null(error)) {
    stop(error)
  } else {
    success
  }
}

expect_promise <- function(p, state = NULL) {
  name <- deparse(substitute(p))
  expect(
    promises::is.promise(p),
    sprintf("`%s` is not a promise", name)
  )
  if (!is.null(state)) {
    expect_equal(attr(p, "promise_impl")$status(), state)
  }
}

# Chat component markup

    Code
      chat_ui("chat")
    Output
      <shiny-chat-container class="html-fill-item html-fill-container" data-require-bs-caller="chat_ui" data-require-bs-version="5" fill id="chat" placeholder="Enter a message..." style="width:min(680px, 100%);height:auto;">
        <shiny-chat-messages></shiny-chat-messages>
        <shiny-chat-input id="chat_user_input" placeholder="Enter a message..."></shiny-chat-input>
      </shiny-chat-container>

---

    Code
      chat_ui("chat", messages = list("Foo", "Bar"))
    Output
      <shiny-chat-container class="html-fill-item html-fill-container" data-require-bs-caller="chat_ui" data-require-bs-version="5" fill id="chat" placeholder="Enter a message..." style="width:min(680px, 100%);height:auto;">
        <shiny-chat-messages>
          <shiny-chat-message data-role="assistant" content="Foo"></shiny-chat-message>
          <shiny-chat-message data-role="assistant" content="Bar"></shiny-chat-message>
        </shiny-chat-messages>
        <shiny-chat-input id="chat_user_input" placeholder="Enter a message..."></shiny-chat-input>
      </shiny-chat-container>

---

    Code
      chat_ui("chat", messages = list(list(content = "Assistant", role = "assistant"),
      list(content = "User", role = "user")))
    Output
      <shiny-chat-container class="html-fill-item html-fill-container" data-require-bs-caller="chat_ui" data-require-bs-version="5" fill id="chat" placeholder="Enter a message..." style="width:min(680px, 100%);height:auto;">
        <shiny-chat-messages>
          <shiny-chat-message data-role="assistant" content="Assistant"></shiny-chat-message>
          <shiny-chat-message data-role="user" content="User"></shiny-chat-message>
        </shiny-chat-messages>
        <shiny-chat-input id="chat_user_input" placeholder="Enter a message..."></shiny-chat-input>
      </shiny-chat-container>

---

    Code
      chat_ui("chat", messages = list(div("Hello"), span("world")))
    Output
      <shiny-chat-container class="html-fill-item html-fill-container" data-require-bs-caller="chat_ui" data-require-bs-version="5" fill id="chat" placeholder="Enter a message..." style="width:min(680px, 100%);height:auto;">
        <shiny-chat-messages>
          <shiny-chat-message data-role="assistant" content="&lt;shinychat-raw-html&gt;&#10;  &lt;div&gt;Hello&lt;/div&gt;&#10;&lt;/shinychat-raw-html&gt;"></shiny-chat-message>
          <shiny-chat-message data-role="assistant" content="&lt;shinychat-raw-html&gt;&#10;  &lt;span&gt;world&lt;/span&gt;&#10;&lt;/shinychat-raw-html&gt;"></shiny-chat-message>
        </shiny-chat-messages>
        <shiny-chat-input id="chat_user_input" placeholder="Enter a message..."></shiny-chat-input>
      </shiny-chat-container>

---

    Code
      render_tags(chat_ui("chat", messages = list(div("Hello", htmlDependency("foo",
        "1.0.0", "")), span("world"))))
    Output
      $deps
      [{"name":"foo","all_files":true},{"name":"shinychat","script":{"src":"shinychat.js","type":"module"},"stylesheet":"shinychat.css","all_files":true},{"name":"bslib-tag-require","script":"tag-require.js","all_files":true},{"name":"htmltools-fill","stylesheet":"fill.css","all_files":true}] 
      
      $html
      <shiny-chat-container class="html-fill-item html-fill-container" data-require-bs-caller="chat_ui" data-require-bs-version="5" fill id="chat" placeholder="Enter a message..." style="width:min(680px, 100%);height:auto;">
        <shiny-chat-messages>
          <shiny-chat-message data-role="assistant" content="&lt;shinychat-raw-html&gt;&#10;  &lt;div&gt;Hello&lt;/div&gt;&#10;&lt;/shinychat-raw-html&gt;"></shiny-chat-message>
          <shiny-chat-message data-role="assistant" content="&lt;shinychat-raw-html&gt;&#10;  &lt;span&gt;world&lt;/span&gt;&#10;&lt;/shinychat-raw-html&gt;"></shiny-chat-message>
        </shiny-chat-messages>
        <shiny-chat-input id="chat_user_input" placeholder="Enter a message..."></shiny-chat-input>
      </shiny-chat-container>
      

---

    Code
      react_tag <- tags$div("react", `data-shinychat-react` = NA)
      chat_ui("chat", messages = list(tagList(tags$div("before"), react_tag, tags$div(
        "after"))))
    Output
      <shiny-chat-container class="html-fill-item html-fill-container" data-require-bs-caller="chat_ui" data-require-bs-version="5" fill id="chat" placeholder="Enter a message..." style="width:min(680px, 100%);height:auto;">
        <shiny-chat-messages>
          <shiny-chat-message data-role="assistant" content="&lt;shinychat-raw-html&gt;&#10;  &lt;div&gt;before&lt;/div&gt;&#10;&lt;/shinychat-raw-html&gt;&#10;&lt;div data-shinychat-react&gt;react&lt;/div&gt;&#10;&lt;shinychat-raw-html&gt;&#10;  &lt;div&gt;after&lt;/div&gt;&#10;&lt;/shinychat-raw-html&gt;"></shiny-chat-message>
        </shiny-chat-messages>
        <shiny-chat-input id="chat_user_input" placeholder="Enter a message..."></shiny-chat-input>
      </shiny-chat-container>

# chat_ui() snapshot for plain string greeting

    Code
      chat_ui("chat", greeting = "## Welcome!")
    Output
      <shiny-chat-container class="html-fill-item html-fill-container" data-require-bs-caller="chat_ui" data-require-bs-version="5" fill greeting="{&quot;content&quot;:&quot;## Welcome!&quot;,&quot;content_type&quot;:&quot;markdown&quot;,&quot;options&quot;:{&quot;dismissible&quot;:true}}" id="chat" placeholder="Enter a message..." style="width:min(680px, 100%);height:auto;">
        <shiny-chat-messages></shiny-chat-messages>
        <shiny-chat-input id="chat_user_input" placeholder="Enter a message..."></shiny-chat-input>
      </shiny-chat-container>

# chat_ui() snapshot for chat_greeting with dismissible=FALSE

    Code
      chat_ui("chat", greeting = chat_greeting("## Hi", dismissible = FALSE))
    Output
      <shiny-chat-container class="html-fill-item html-fill-container" data-require-bs-caller="chat_ui" data-require-bs-version="5" fill greeting="{&quot;content&quot;:&quot;## Hi&quot;,&quot;content_type&quot;:&quot;markdown&quot;,&quot;options&quot;:{&quot;dismissible&quot;:false}}" id="chat" placeholder="Enter a message..." style="width:min(680px, 100%);height:auto;">
        <shiny-chat-messages></shiny-chat-messages>
        <shiny-chat-input id="chat_user_input" placeholder="Enter a message..."></shiny-chat-input>
      </shiny-chat-container>

# Chat component markup

    Code
      output_markdown_stream("stream")
    Output
      <shiny-markdown-stream id="stream" style="width:min(680px, 100%);height:auto;margin:0 auto;" content="" content-type="markdown" auto-scroll=""></shiny-markdown-stream>

---

    Code
      output_markdown_stream("stream", content = "Foo\nBar")
    Output
      <shiny-markdown-stream id="stream" style="width:min(680px, 100%);height:auto;margin:0 auto;" content="Foo&#10;Bar" content-type="markdown" auto-scroll=""></shiny-markdown-stream>

---

    Code
      render_tags(output_markdown_stream("stream", content = div("Hello",
        htmlDependency("foo", "1.0.0", ""))))
    Output
      $deps
      [{"name":"foo","all_files":true},{"name":"shinychat","script":{"src":"shinychat.js","type":"module"},"stylesheet":"shinychat.css","all_files":true}] 
      
      $html
      <shiny-markdown-stream id="stream" style="width:min(680px, 100%);height:auto;margin:0 auto;" content="&lt;shinychat-raw-html&gt;&#10;  &lt;div&gt;Hello&lt;/div&gt;&#10;&lt;/shinychat-raw-html&gt;" content-type="markdown" auto-scroll=""></shiny-markdown-stream>
      

# opt_shinychat_tool_display handles options and environment variables

    Code
      opt_shinychat_tool_display()
    Condition
      Error in `opt_shinychat_tool_display()`:
      ! `SHINYCHAT_TOOL_DISPLAY` must be one of "none", "basic", or "rich", not "invalid".

---

    Code
      opt_shinychat_tool_display()
    Condition
      Error in `opt_shinychat_tool_display()`:
      ! `shinychat.tool_display` must be one of "none", "basic", or "rich", not "invalid".

# ContentToolResult requires an associated `@request` property

    Code
      contents_shinychat(new_tool_result(request = NULL))
    Condition
      Error in `method(contents_shinychat, ellmer::ContentToolResult)`:
      ! `ContentToolResult` objects must have an associated `@request` property.

# get_tool_result_display handles invalid formats

    Code
      get_tool_result_display(result)
    Condition
      Warning:
      Invalid `@extra$display` format for `ContentToolResult` from `test-tool()` (call id: test-id).
      i To display HTML content for tool results in shinychat, create a tool result with `extra = list(display = list(html = ...))`.
      i You can also use `markdown` or `text` items in `display` to show Markdown or plain text, respectively.
    Output
      list()

---

    Code
      get_tool_result_display(result)
    Condition
      Warning:
      Invalid `@extra$display` format for `ContentToolResult` from `test-tool()` (call id: test-id).
      x Expected a list with fields `html`, `markdown`, `text`, `show_request`, `open`, `full_screen`, `title`, `icon`, or `footer`, not a string.
    Output
      list()

# throws when a result does not have a `request` property

    Code
      contents_shinychat(new_tool_result(request = NULL))
    Condition
      Error in `method(contents_shinychat, ellmer::ContentToolResult)`:
      ! `ContentToolResult` objects must have an associated `@request` property.

# throws for invalid tool display option

    Code
      opt_shinychat_tool_display()
    Condition
      Error in `opt_shinychat_tool_display()`:
      ! `shinychat.tool_display` must be one of "none", "basic", or "rich", not "invalid".

# throws for invalid tool display ennvar

    Code
      opt_shinychat_tool_display()
    Condition
      Error in `opt_shinychat_tool_display()`:
      ! `SHINYCHAT_TOOL_DISPLAY` must be one of "none", "basic", or "rich", not "invalid".

# warns when `display` is not a list

    Code
      as.tags(contents_shinychat(result))
    Condition
      Warning:
      Invalid `@extra$display` format for `ContentToolResult` from `test-tool()` (call id: test-id).
      i To display HTML content for tool results in shinychat, create a tool result with `extra = list(display = list(html = ...))`.
      i You can also use `markdown` or `text` items in `display` to show Markdown or plain text, respectively.
    Output
      <shiny-tool-result request-id="test-id" tool-name="test-tool" request-call="`test-tool`()" status="success" show-request value="{}" value-type="code" data-shinychat-react></shiny-tool-result>

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{chat_ui}
\alias{chat_ui}
\title{Create a chat UI element}
\usage{
chat_ui(
  id,
  ...,
  messages = NULL,
  greeting = NULL,
  placeholder = "Enter a message...",
  width = "min(680px, 100\%)",
  height = "auto",
  fill = TRUE,
  icon_assistant = NULL,
  enable_cancel = FALSE,
  footer = NULL
)
}
\arguments{
\item{id}{The ID of the chat element}

\item{...}{Extra HTML attributes to include on the chat element}

\item{messages}{A list of messages to prepopulate the chat with. Each
message can be one of the following:
\itemize{
\item A string, which is interpreted as markdown and rendered to HTML on
the client.
\itemize{
\item To prevent interpreting as markdown, mark the string as
\code{\link[htmltools:HTML]{htmltools::HTML()}}.
}
\item A UI element.
\itemize{
\item This includes \code{\link[htmltools:tagList]{htmltools::tagList()}}, which take UI elements
(including strings) as children. In this case, strings are still
interpreted as markdown as long as they're not inside HTML.
}
\item A named list of \code{content} and \code{role}. The \code{content} can contain content
as described above, and the \code{role} can be "assistant" or "user".
}}

\item{greeting}{An optional greeting to display when the chat first loads.
Can be a \code{\link[=chat_greeting]{chat_greeting()}} object, or a plain string (which is
auto-wrapped with default options). The greeting is dismissed when the
user sends their first message. For example:
\code{greeting = chat_greeting("## Hello!\\n\\nHow can I help you today?")}}

\item{placeholder}{The placeholder text for the chat's user input field}

\item{width}{The CSS width of the chat element}

\item{height}{The CSS height of the chat element}

\item{fill}{Whether the chat element should try to vertically fill its
container, if the container is
\href{https://rstudio.github.io/bslib/articles/filling/index.html}{fillable}}

\item{icon_assistant}{The icon to use for the assistant chat messages.
Can be HTML or a tag in the form of \code{\link[htmltools:HTML]{htmltools::HTML()}} or
\code{\link[htmltools:tags]{htmltools::tags()}}. If \code{None}, a default robot icon is used.}

\item{enable_cancel}{If \code{TRUE}, show a stop button during streaming that
allows the user to cancel the in-progress response. When using
\code{\link[=chat_mod_server]{chat_mod_server()}}, cancellation is wired up automatically. For manual
usage with \code{chat_ui()}, observe \verb{input$<id>_cancel} to handle cancellation
(e.g., by calling \code{ctrl$cancel()} on an ellmer \code{stream_controller()}).
Defaults to \code{FALSE}.}

\item{footer}{Optional HTML content to display below the chat input.
This can be any HTML content (tags, tag lists, or character strings).
Useful for adding disclaimers, attribution, or other information.
The footer text is styled slightly smaller and lighter than body text
by default. Customize with CSS properties \code{--shiny-chat-footer-font-size}
and \code{--shiny-chat-footer-color} on the chat container or footer element.}
}
\value{
A Shiny tag object, suitable for inclusion in a Shiny UI
}
\description{
Inserts a chat UI element into a Shiny UI, which includes a scrollable
section for displaying chat messages, and an input field for the user to
enter new messages.

To respond to user input, listen for \code{input$ID_user_input} (for example, if
\code{id="my_chat"}, user input will be at \code{input$my_chat_user_input}), and use
\code{\link[=chat_append]{chat_append()}} to append messages to the chat.
}
\section{Greeting}{


A greeting is an optional welcome message shown before any conversation
messages. It is automatically dismissed when the user sends their first
message (unless created with \code{dismissible = FALSE}).

\strong{Static greeting.} Pass a string or \code{\link[=chat_greeting]{chat_greeting()}} to the \code{greeting}
parameter:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{chat_ui("chat", greeting = "## Hello!\\n\\nHow can I help you today?")
}\if{html}{\out{</div>}}

\strong{Dynamic greeting from the server.} Leave \code{greeting} unset and use
\code{\link[=chat_set_greeting]{chat_set_greeting()}} from your server function. This is useful when the
greeting depends on session state or is generated by a model.

\strong{\code{greeting_requested} input.} When the chat is visible on the page, has
no messages, and has no greeting set, Shiny fires
\verb{input$<id>_greeting_requested} (e.g. \code{input$chat_greeting_requested} for
\code{chat_ui("chat")}). The value is an event counter suitable for
\code{\link[shiny:observeEvent]{shiny::observeEvent()}}. Use it to trigger server-side greeting generation:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{observeEvent(input$chat_greeting_requested, \{
  stream <- chat_client$stream_async("Generate a short welcome message.")
  chat_set_greeting("chat", chat_greeting(stream))
\})
}\if{html}{\out{</div>}}

This input fires when the chat component is first viewed on the page and
empty, and again after \code{\link[=chat_clear]{chat_clear()}} \code{(greeting = TRUE)}, enabling a
regenerate pattern where clearing the greeting automatically triggers a
fresh one.
}

\section{Thinking display}{


When a model produces reasoning or "thinking" tokens, shinychat renders them
in a collapsible panel above the response. The panel shows a live stream of
the model's reasoning while it thinks, then auto-collapses when the response
begins.

Thinking display works automatically with any model that supports it. Two
paths are supported:
\enumerate{
\item \strong{ellmer's \code{ContentThinking} objects.} Models that provide a structured
thinking API (e.g., Claude with extended thinking) emit \code{ContentThinking}
objects when you stream with \code{stream = "content"}. shinychat detects these
and routes them to the thinking panel. This is what \code{chat_append()} uses
internally when you pass it an ellmer content stream.
\item \strong{Raw \verb{<thinking>} tags.} Many open-source and local models (DeepSeek,
QwQ, Qwen, etc.) emit \verb{<thinking>...</thinking>} tags directly in their
markdown output. shinychat detects these tags during streaming and renders
the enclosed text in the thinking panel with no extra configuration.
}
\subsection{Topic labels}{

You can optionally get labeled sub-sections within the thinking panel by
asking the model to emit \verb{<topic>...</topic>} tags in its reasoning. These
are extracted and rendered as section headings inside the thinking display,
and the current topic appears in the collapsed header as a live status.

To use topic labels, add something like this to your system prompt:

\if{html}{\out{<div class="sourceCode">}}\preformatted{When thinking through a problem, wrap brief topic labels in <topic> tags
to indicate what you're currently reasoning about. For example:
<topic>parsing the input</topic>
}\if{html}{\out{</div>}}

Topic labels are entirely optional. Without them, the thinking panel still
works -- it just won't have sub-section headings.
}
}

\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)
library(shinychat)

ui <- page_fillable(
  chat_ui("chat", fill = TRUE)
)

server <- function(input, output, session) {
  observeEvent(input$chat_user_input, {
    # In a real app, this would call out to a chat client or API,
    # perhaps using the 'ellmer' package.
    response <- paste0(
      "You said:\n\n",
      "<blockquote>",
      htmltools::htmlEscape(input$chat_user_input),
      "</blockquote>"
    )
    chat_append("chat", response)
    chat_append("chat", stream)
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat_restore.R
\name{chat_restore}
\alias{chat_restore}
\title{Add Shiny bookmarking for shinychat}
\usage{
chat_restore(
  id,
  client,
  ...,
  bookmark_on_input = TRUE,
  bookmark_on_response = TRUE,
  restore_ui = TRUE,
  session = getDefaultReactiveDomain()
)
}
\arguments{
\item{id}{The ID of the chat element}

\item{client}{The \pkg{ellmer} LLM chat client.}

\item{...}{Used for future parameter expansion.}

\item{bookmark_on_input}{A logical value determines if the bookmark should be updated when the user submits a message. Default is \code{TRUE}.}

\item{bookmark_on_response}{A logical value determines if the bookmark should be updated when the response stream completes. Default is \code{TRUE}.}

\item{restore_ui}{Whether to render the client's existing turns into the
chat UI on registration. Default is \code{TRUE}. Set to \code{FALSE} when
re-registering bookmarks after a client swap (where the UI already reflects
the conversation).}

\item{session}{The Shiny session object}
}
\value{
Invisibly returns a function that, when called, cancels all
bookmark registrations made by this call. This is useful when swapping
the chat client: cancel the previous bookmarks, then call
\code{chat_restore()} again with the new client.
}
\description{
Adds Shiny bookmarking hooks to save and restore the \pkg{ellmer} chat
\code{client}. Also restores chat messages from the history in the \code{client}.

If either \code{bookmark_on_input} or \code{bookmark_on_response} is \code{TRUE}, the Shiny
App's bookmark will be automatically updated without showing a modal to the
user.

Note: Only the \code{client}'s chat state is saved/restored in the bookmark. If
the \code{client}'s state doesn't properly capture the chat's UI (i.e., a
transformation is applied in-between receiving and displaying the message),
then you may need to implement your own \code{session$onRestore()} (and possibly
\code{session$onBookmark}) handler to restore any additional state.

To avoid restoring chat history from the \code{client}, you can ensure that the
history is empty by calling \code{client$set_turns(list())} before passing the
client to \code{chat_restore()}.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)
library(shinychat)

ui <- function(request) {
  page_fillable(
    chat_ui("chat", fill = TRUE)
  )
}

server <- function(input, output, session) {
  chat_client <- ellmer::chat_ollama(
    system_prompt = "Important: Always respond in a limerick",
    model = "qwen2.5-coder:1.5b",
    echo = TRUE
  )
  # Update bookmark to chat on user submission and completed response
  chat_restore("chat", chat_client)

  observeEvent(input$chat_user_input, {
    stream <- chat_client$stream_async(input$chat_user_input)
    chat_append("chat", stream)
  })
}

# Enable bookmarking!
shinyApp(ui, server, enableBookmarking = "server")
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown-stream.R
\name{output_markdown_stream}
\alias{output_markdown_stream}
\title{Create a UI element for a markdown stream.}
\usage{
output_markdown_stream(
  id,
  ...,
  content = "",
  content_type = "markdown",
  auto_scroll = TRUE,
  width = "min(680px, 100\%)",
  height = "auto"
)
}
\arguments{
\item{id}{A unique identifier for this markdown stream.}

\item{...}{Extra HTML attributes to include on the chat element}

\item{content}{A string of content to display before any streaming occurs.
When \code{content_type} is Markdown or HTML, it may also be UI element(s) such
as input and output bindings.}

\item{content_type}{The content type. Default is \code{"markdown"} (specifically,
CommonMark). Supported content types include:
* \code{"markdown"}: markdown text, specifically CommonMark
* \code{"html"}: for rendering HTML content.
* \code{"text"}: for plain text.}

\item{auto_scroll}{Whether to automatically scroll to the bottom of a
scrollable container when new content is added. Default is True.}

\item{width}{The width of the UI element.}

\item{height}{The height of the UI element.}
}
\value{
A shiny tag object.
}
\description{
Creates a UI element for a \code{\link[=markdown_stream]{markdown_stream()}}. A markdown stream can be
useful for displaying generative AI responses (outside of a chat interface),
streaming logs, or other use cases where chunks of content are generated
over time.
}
\seealso{
\code{\link[=markdown_stream]{markdown_stream()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{chat_append}
\alias{chat_append}
\title{Append an assistant response (or user message) to a chat control}
\usage{
chat_append(
  id,
  response,
  role = c("assistant", "user"),
  icon = NULL,
  session = getDefaultReactiveDomain()
)
}
\arguments{
\item{id}{The ID of the chat element}

\item{response}{The message or message stream to append to the chat element.
The actual message content can one of the following:
\itemize{
\item A string, which is interpreted as markdown and rendered to HTML on
the client.
\itemize{
\item To prevent interpreting as markdown, mark the string as
\code{\link[htmltools:HTML]{htmltools::HTML()}}.
}
\item A UI element.
\itemize{
\item This includes \code{\link[htmltools:tagList]{htmltools::tagList()}}, which take UI elements
(including strings) as children. In this case, strings are still
interpreted as markdown as long as they're not inside HTML.
}
}}

\item{role}{The role of the message (either "assistant" or "user"). Defaults
to "assistant".}

\item{icon}{An optional icon to display next to the message, currently only
used for assistant messages. The icon can be any HTML element (e.g., an
\code{\link[htmltools:img]{htmltools::img()}} tag) or a string of HTML.}

\item{session}{The Shiny session object}
}
\value{
Returns a promise that resolves to the contents of the stream, or an
error. This promise resolves when the message has been successfully sent to
the client; note that it does not guarantee that the message was actually
received or rendered by the client. The promise rejects if an error occurs
while processing the response (see the "Error handling" section).
}
\description{
The \code{chat_append} function appends a message to an existing \code{\link[=chat_ui]{chat_ui()}}. The
\code{response} can be a string, string generator, string promise, or string
promise generator (as returned by the 'ellmer' package's \code{chat}, \code{stream},
\code{chat_async}, and \code{stream_async} methods, respectively).

This function should be called from a Shiny app's server. It is generally
used to append the client's response to the chat, while user messages are
added to the chat UI automatically by the front-end. You'd only need to use
\code{chat_append(role="user")} if you are programmatically generating queries
from the server and sending them on behalf of the user, and want them to be
reflected in the UI.
}
\section{Error handling}{
If the \code{response} argument is a generator, promise, or promise generator, and
an error occurs while producing the message (e.g., an iteration in
\code{stream_async} fails), the promise returned by \code{chat_append} will reject with
the error. If the \code{chat_append} call is the last expression in a Shiny
observer, shinychat will log the error message and show a message that the
error occurred in the chat UI.
}

\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(coro)
library(bslib)
library(shinychat)

# Dumbest chatbot in the world: ignores user input and chooses
# a random, vague response.
fake_chatbot <- async_generator(function(input) {
  responses <- c(
    "What does that suggest to you?",
    "I see.",
    "I'm not sure I understand you fully.",
    "What do you think?",
    "Can you elaborate on that?",
    "Interesting question! Let's examine thi... **See more**"
  )

  await(async_sleep(1))
  for (chunk in strsplit(sample(responses, 1), "")[[1]]) {
    yield(chunk)
    await(async_sleep(0.02))
  }
})

ui <- page_fillable(
  chat_ui("chat", fill = TRUE)
)

server <- function(input, output, session) {
  observeEvent(input$chat_user_input, {
    response <- fake_chatbot(input$chat_user_input)
    chat_append("chat", response)
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/markdown-stream.R
\name{markdown_stream}
\alias{markdown_stream}
\title{Stream markdown content}
\usage{
markdown_stream(
  id,
  content_stream,
  operation = c("replace", "append"),
  session = getDefaultReactiveDomain()
)
}
\arguments{
\item{id}{The ID of the markdown stream to stream content to.}

\item{content_stream}{A string generator (e.g., \code{\link[coro:generator]{coro::generator()}} or
\code{\link[coro:async_generator]{coro::async_generator()}}), a string promise (e.g., \code{\link[promises:promise]{promises::promise()}}),
or a string promise generator.}

\item{operation}{The operation to perform on the markdown stream. The default,
\code{"replace"}, will replace the current content with the new content stream.
The other option, \code{"append"}, will append the new content stream to the
existing content.}

\item{session}{The Shiny session object.}
}
\description{
Streams markdown content into a \code{\link[=output_markdown_stream]{output_markdown_stream()}} UI element.  A
markdown stream can be useful for displaying generative AI responses (outside
of a chat interface), streaming logs, or other use cases where chunks of
content are generated over time.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}

library(shiny)
library(coro)
library(bslib)
library(shinychat)

# Define a generator that yields a random response
# (imagine this is a more sophisticated AI generator)
random_response_generator <- async_generator(function() {
  responses <- c(
    "What does that suggest to you?",
    "I see.",
    "I'm not sure I understand you fully.",
    "What do you think?",
    "Can you elaborate on that?",
    "Interesting question! Let's examine thi... **See more**"
  )

  await(async_sleep(1))
  for (chunk in strsplit(sample(responses, 1), "")[[1]]) {
    yield(chunk)
    await(async_sleep(0.02))
  }
})

ui <- page_fillable(
  actionButton("generate", "Generate response"),
  output_markdown_stream("stream")
)

server <- function(input, output, session) {
  observeEvent(input$generate, {
    markdown_stream("stream", random_response_generator())
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat_app.R
\name{chat_app}
\alias{chat_app}
\alias{chat_mod_ui}
\alias{chat_mod_server}
\title{Open a live chat application in the browser}
\usage{
chat_app(client, ..., bookmark_store = "url")

chat_mod_ui(id, ..., client = deprecated(), messages = NULL)

chat_mod_server(
  id,
  client,
  greeting = NULL,
  bookmark_on_input = TRUE,
  bookmark_on_response = TRUE
)
}
\arguments{
\item{client}{A chat object created by \pkg{ellmer}, e.g.
\code{\link[ellmer:chat_openai]{ellmer::chat_openai()}} and friends. This argument is deprecated in
\code{chat_mod_ui()} because the client state is now managed by
\code{chat_mod_server()}.}

\item{...}{In \code{chat_app()}, additional arguments are passed to
\code{\link[shiny:shinyApp]{shiny::shinyApp()}}. In \code{chat_mod_ui()}, additional arguments are passed to
\code{\link[=chat_ui]{chat_ui()}}.}

\item{bookmark_store}{The bookmarking store to use for the app. Passed to
\code{enable_bookmarking} in \code{\link[shiny:shinyApp]{shiny::shinyApp()}}. Defaults to \code{"url"}, which
uses the URL to store the chat state. URL-based bookmarking is limited in
size; use \code{"server"} to store the state on the server side without size
limitations; or disable bookmarking by setting this to \code{"disable"}.}

\item{id}{The chat module ID.}

\item{messages}{Initial messages shown in the chat, used only when \code{client}
(in \code{chat_mod_ui()}) doesn't already contain turns. Passed to \code{messages}
in \code{\link[=chat_ui]{chat_ui()}}.}

\item{greeting}{Optional greeting to set when the module initializes.
Accepts a static value (string, \code{\link[htmltools:HTML]{htmltools::HTML()}}, \code{\link[htmltools:tagList]{htmltools::tagList()}},
or \code{\link[=chat_greeting]{chat_greeting()}}) or a \strong{function} that generates the greeting
dynamically. See the \strong{Greeting} section below for details.}

\item{bookmark_on_input}{A logical value determines if the bookmark should be updated when the user submits a message. Default is \code{TRUE}.}

\item{bookmark_on_response}{A logical value determines if the bookmark should be updated when the response stream completes. Default is \code{TRUE}.}
}
\value{
\itemize{
\item \code{chat_app()} returns a \code{\link[shiny:shinyApp]{shiny::shinyApp()}} object.
\item \code{chat_mod_ui()} returns the UI for a shinychat module.
\item \code{chat_mod_server()} includes the shinychat module server logic, and
returns an environment containing:
\itemize{
\item \code{last_input}: A reactive value containing the last user input.
\item \code{last_turn}: A reactive value containing the last assistant turn.
\item \code{update_user_input()}: A function to update the chat input or submit a
new user input. Takes the same arguments as \code{\link[=update_chat_user_input]{update_chat_user_input()}},
except for \code{id} and \code{session}, which are supplied automatically.
\item \code{append()}: A function to append a new message to the chat UI. Takes
the same arguments as \code{\link[=chat_append]{chat_append()}}, except for \code{id} and \code{session},
which are supplied automatically.
\item \code{clear()}: A function to clear the chat history and the chat UI.
\code{clear()} takes an optional list of \code{messages} used to initialize the
chat after clearing. \code{messages} should be a list of messages, where
each message is a list with \code{role} and \code{content} fields. The
\code{client_history} argument controls how the chat client's history is
updated after clearing. It can be one of: \code{"clear"} the chat history;
\code{"set"} the chat history to \code{messages}; \code{"append"} \code{messages} to the
existing chat history; or \code{"keep"} the existing chat history.
\item \code{set_greeting()}: A function to set, stream, or clear the chat
greeting. Pass a \code{\link[=chat_greeting]{chat_greeting()}} object, a plain string, or
\code{NULL} to clear. Streaming greetings run inside an
\link[shiny:ExtendedTask]{shiny::ExtendedTask} so the session stays responsive; if called
while a greeting is already streaming, the new greeting is queued.
If the greeting has already been dismissed, calling \code{set_greeting()}
updates the content but does not make it visible again; call
\code{clear(greeting = TRUE)} first to show a new greeting after dismissal.
\item \code{status}: A reactive value indicating the current chat interaction
state. Returns \code{"idle"} when no response is in progress, or
\code{"streaming"} while a response is actively being received.
\item \code{client}: The current chat client object (an active binding that
always reflects the latest client, even after \code{set_client()}
is called).
\item \code{set_client(new_client, sync = TRUE)}: Replace the chat client used by
the module. When \code{sync} is \code{TRUE} (the default), the new client
inherits conversation turns, system prompt, and tools from the previous
client so the conversation continues seamlessly. Set \code{sync = FALSE} to
use the new client as-is. If a response is currently streaming, the
swap is deferred until the stream completes. If called multiple times
while streaming, only the most recent new client is used.
}
}
}
\description{
Create a simple Shiny app for live chatting using an \link[ellmer:Chat]{ellmer::Chat} object.
Note that these functions will mutate the input \code{client} object as
you chat because your turns will be appended to the history.

The app created by \code{chat_app()} is suitable for interactive use by a single
user. For multi-user Shiny apps, use the Shiny module chat functions --
\code{chat_mod_ui()} and \code{chat_mod_server()} -- and be sure to create a new chat
client for each user session.
}
\section{Functions}{
\itemize{
\item \code{chat_app()}: A simple Shiny app for live chatting. Note that this
app is suitable for interactive use by a single user; do not use
\code{chat_app()} in a multi-user Shiny app context.

\item \code{chat_mod_ui()}: A simple chat app module UI.

\item \code{chat_mod_server()}: A simple chat app module server.

}}
\section{Greeting}{


When \code{greeting} is a \strong{function}, the module calls it each time the
\code{greeting_requested} event fires — on first view when the chat is empty,
and again after \code{clear(greeting = TRUE)}. The function should return a
\code{\link[=chat_greeting]{chat_greeting()}} (typically wrapping a stream). Static values (strings,
\code{\link[=chat_greeting]{chat_greeting()}} objects) are set once at init and do not regenerate.

The module detects \strong{named arguments} in the greeting function to decide
what to pass. Currently the only recognized argument is \code{client}.

\strong{\verb{function(client)}} (recommended). The module clones the \code{client}
passed to \code{chat_mod_server()}, wipes its turn history, and passes the
fresh clone as \code{client}. This avoids manually creating and configuring
a separate client:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{chat_mod_server("chat", client, greeting = function(client) \{
  stream <- client$stream_async("Generate a short welcome message.")
  chat_greeting(stream)
\})
}\if{html}{\out{</div>}}

\strong{\verb{function()}} (zero arguments). You create and manage your own client:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{chat_mod_server("chat", client, greeting = function() \{
  greeter <- ellmer::chat_openai(model = "gpt-4o")
  stream <- greeter$stream_async("Generate a short welcome message.")
  chat_greeting(stream)
\})
}\if{html}{\out{</div>}}

\strong{Static value.} Set once; does not regenerate after \code{clear()}:

\if{html}{\out{<div class="sourceCode r">}}\preformatted{chat_mod_server("chat", client, greeting = "## Welcome!\\n\\nHow can I help?")
}\if{html}{\out{</div>}}

The returned \code{set_greeting()} helper is available for cases where you need
to set a greeting outside the greeting lifecycle.
}

\examples{
\dontrun{
# Interactive in the console ----
client <- ellmer::chat_anthropic()
chat_app(client)

# Inside a Shiny app ----
library(shiny)
library(bslib)
library(shinychat)

ui <- page_fillable(
  titlePanel("shinychat example"),

  layout_columns(
    card(
      card_header("Chat with Claude"),
      chat_mod_ui(
        "claude",
        messages = list(
          "Hi! Use this chat interface to chat with Anthropic's `claude-3-5-sonnet`."
        )
      )
    ),
    card(
      card_header("Chat with ChatGPT"),
      chat_mod_ui(
        "openai",
        messages = list(
          "Hi! Use this chat interface to chat with OpenAI's `gpt-4o`."
        )
      )
    )
  )
)

server <- function(input, output, session) {
  claude <- ellmer::chat_anthropic(model = "claude-3-5-sonnet-latest") # Requires ANTHROPIC_API_KEY
  openai <- ellmer::chat_openai(model = "gpt-4o") # Requires OPENAI_API_KEY

  chat_mod_server("claude", claude)
  chat_mod_server("openai", openai)
}

shinyApp(ui, server)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{chat_set_greeting}
\alias{chat_set_greeting}
\title{Set the greeting for a chat UI}
\usage{
chat_set_greeting(id, greeting, session = getDefaultReactiveDomain())
}
\arguments{
\item{id}{The ID of the chat element}

\item{greeting}{The greeting to display. Can be:
\itemize{
\item \code{NULL} to clear the current greeting.
\item A plain string or \code{\link[htmltools:HTML]{htmltools::HTML()}} object (auto-wrapped with default
options).
\item A \code{\link[=chat_greeting]{chat_greeting()}} object.
}}

\item{session}{The Shiny session object}
}
\value{
Returns \code{invisible(NULL)} for static and NULL greetings, or a
promise for streaming greetings (resolves when streaming is complete).
}
\description{
Sets or clears the greeting displayed in an existing \code{\link[=chat_ui]{chat_ui()}}. The
greeting is shown when the chat first loads and is dismissed when the user
sends their first message.

Call \code{chat_set_greeting()} from the server to display a dynamic or streaming
greeting. This is typically used inside an \code{\link[shiny:observeEvent]{shiny::observeEvent()}} watching
\verb{input$<id>_greeting_requested} -- an event that fires when the chat is
visible, has no messages, and has no greeting set. See the \strong{Greeting}
section in \code{\link[=chat_ui]{chat_ui()}} for details on \code{greeting_requested}.

If the greeting has already been dismissed, calling this function updates
the greeting content but does not make it visible again. To show a new
greeting after dismissal, first clear the chat with
\code{chat_clear(id, greeting = TRUE)}.

Pass \code{NULL} to clear the current greeting entirely.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)
library(shinychat)

# Static greeting set from the server
ui <- page_fillable(chat_ui("chat"))

server <- function(input, output, session) {
  chat_set_greeting("chat", "## Welcome!\n\nHow can I help you today?")

  observeEvent(input$chat_user_input, {
    chat_append("chat", paste0("You said: ", input$chat_user_input))
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)
library(shinychat)
library(coro)

# Streaming greeting generated by a function
greeting_generator <- async_generator(function() {
  for (chunk in strsplit("## Hello!\n\nHow can I help you?", "")[[1]]) {
    yield(chunk)
    await(async_sleep(0.02))
  }
})

ui <- page_fillable(chat_ui("chat"))

server <- function(input, output, session) {
  chat_set_greeting("chat", chat_greeting(greeting_generator()))

  observeEvent(input$chat_user_input, {
    chat_append("chat", paste0("You said: ", input$chat_user_input))
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)
library(shinychat)

# LLM-generated greeting using greeting_requested
ui <- page_fillable(chat_ui("chat"))

server <- function(input, output, session) {
  chat_client <- ellmer::chat_openai(model = "gpt-4o")

  observeEvent(input$chat_greeting_requested, {
    stream <- chat_client$stream_async(
      "Generate a short, friendly welcome message."
    )
    chat_set_greeting("chat", chat_greeting(stream))
  })

  observeEvent(input$chat_user_input, {
    stream <- chat_client$stream_async(input$chat_user_input)
    chat_append("chat", stream)
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)
library(shinychat)

# Regenerate pattern: chat_clear(greeting = TRUE) triggers greeting_requested
ui <- page_fillable(
  chat_ui("chat"),
  actionButton("regenerate", "New greeting")
)

server <- function(input, output, session) {
  observeEvent(input$chat_greeting_requested, {
    chat_set_greeting(
      "chat",
      paste("## Welcome!\n\nGenerated at", Sys.time())
    )
  })

  observeEvent(input$regenerate, {
    chat_clear("chat", greeting = TRUE)
  })

  observeEvent(input$chat_user_input, {
    chat_append("chat", paste0("You said: ", input$chat_user_input))
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{update_chat_user_input}
\alias{update_chat_user_input}
\title{Update the user input of a chat control}
\usage{
update_chat_user_input(
  id,
  ...,
  value = NULL,
  placeholder = NULL,
  submit = FALSE,
  focus = FALSE,
  session = getDefaultReactiveDomain()
)
}
\arguments{
\item{id}{The ID of the chat element}

\item{...}{Currently unused, but reserved for future use.}

\item{value}{The value to set the user input to. If \code{NULL}, the input will not be updated.}

\item{placeholder}{The placeholder text for the user input}

\item{submit}{Whether to automatically submit the text for the user. Requires \code{value}.}

\item{focus}{Whether to move focus to the input element. Requires \code{value}.}

\item{session}{The Shiny session object}
}
\description{
Update the user input of a chat control
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)
library(shinychat)

ui <- page_fillable(
  chat_ui("chat"),
  layout_columns(
    fill = FALSE,
    actionButton("update_placeholder", "Update placeholder"),
    actionButton("update_value", "Update user input")
  )
)

server <- function(input, output, session) {
  observeEvent(input$update_placeholder, {
    update_chat_user_input("chat", placeholder = "New placeholder text")
  })

  observeEvent(input$update_value, {
    update_chat_user_input("chat", value = "New user input", focus = TRUE)
  })

  observeEvent(input$chat_user_input, {
    response <- paste0("You said: ", input$chat_user_input)
    chat_append("chat", response)
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{chat_append_message}
\alias{chat_append_message}
\title{Low-level function to append a message to a chat control}
\usage{
chat_append_message(
  id,
  msg,
  chunk = TRUE,
  operation = c("append", "replace"),
  icon = NULL,
  session = getDefaultReactiveDomain()
)
}
\arguments{
\item{id}{The ID of the chat element}

\item{msg}{The message to append. Should be a named list with \code{role} and
\code{content} fields. The \code{role} field should be either "user" or "assistant".
The \code{content} field should be a string containing the message content, in
Markdown format.}

\item{chunk}{Whether \code{msg} is just a chunk of a message, and if so, what
type. If \code{FALSE}, then \code{msg} is a complete message. If \code{"start"}, then
\code{msg} is the first chunk of a multi-chunk message. If \code{"end"}, then \code{msg}
is the last chunk of a multi-chunk message. If \code{TRUE}, then \code{msg} is an
intermediate chunk of a multi-chunk message. Default is \code{FALSE}.}

\item{operation}{The operation to perform on the message. If \code{"append"},
then the new content is appended to the existing message content. If
\code{"replace"}, then the existing message content is replaced by the new
content. Ignored if \code{chunk} is \code{FALSE}.}

\item{icon}{An optional icon to display next to the message, currently only
used for assistant messages. The icon can be any HTML element (e.g.,
\code{\link[htmltools:img]{htmltools::img()}} tag) or a string of HTML.}

\item{session}{The Shiny session object}
}
\value{
Returns nothing (\code{invisible(NULL)}).
}
\description{
For advanced users who want to control the message chunking behavior. Most
users should use \code{\link[=chat_append]{chat_append()}} instead.
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(coro)
library(bslib)
library(shinychat)

# Dumbest chatbot in the world: ignores user input and chooses
# a random, vague response.
fake_chatbot <- async_generator(function(id, input) {
  responses <- c(
    "What does that suggest to you?",
    "I see.",
    "I'm not sure I understand you fully.",
    "What do you think?",
    "Can you elaborate on that?",
    "Interesting question! Let's examine thi... **See more**"
  )

  # Use low-level chat_append_message() to temporarily set a progress message
  chat_append_message(id, list(role = "assistant", content = "_Thinking..._ "))
  await(async_sleep(1))
  # Clear the progress message
  chat_append_message(id, list(role = "assistant", content = ""), operation = "replace")

  for (chunk in strsplit(sample(responses, 1), "")[[1]]) {
    yield(chunk)
    await(async_sleep(0.02))
  }
})

ui <- page_fillable(
  chat_ui("chat", fill = TRUE)
)

server <- function(input, output, session) {
  observeEvent(input$chat_user_input, {
    response <- fake_chatbot("chat", input$chat_user_input)
    chat_append("chat", response)
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/contents_shinychat.R
\name{contents_shinychat}
\alias{contents_shinychat}
\title{Format ellmer content for shinychat}
\usage{
contents_shinychat(content)
}
\arguments{
\item{content}{An \code{\link[ellmer:Content]{ellmer::Content}} object.}
}
\value{
Returns text, HTML, or web component tags formatted for use in
\code{chat_ui()}.
}
\description{
Format ellmer content for shinychat
}
\section{Extending \code{contents_shinychat()}}{


You can extend \code{contents_shinychat()} to handle custom content types in your
application. \code{contents_shinychat()} is \link[S7:new_generic]{an S7 generic}. If
you haven't worked with S7 before, you can learn more about S7 classes,
generics and methods in the \href{https://rconsortium.github.io/S7/articles/S7.html}{S7 documentation}.

We'll work through a short example creating a custom display for the results
of a tool that gets local weather forecasts. We first need to create a custom
class that extends \link[ellmer:ContentToolResult]{ellmer::ContentToolResult}.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{library(ellmer)

WeatherToolResult <- S7::new_class(
  "WeatherToolResult",
  parent = ContentToolResult,
  properties = list(
    location_name = S7::class_character
  )
)
}\if{html}{\out{</div>}}

Next, we'll create a simple \code{\link[ellmer:tool]{ellmer::tool()}} that gets the weather forecast
for a location and returns our custom \code{WeatherToolResult} class. The custom
class works just like a regular \code{ContentToolResult}, but it has an additional
\code{location_name} property.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{get_weather_forecast <- tool(
  function(lat, lon, location_name) \{
    WeatherToolResult(
      weathR::point_tomorrow(lat, lon, short = FALSE),
      location_name = location_name
    )
  \},
  name = "get_weather_forecast",
  description = "Get the weather forecast for a location.",
  arguments = list(
    lat = type_number("Latitude"),
    lon = type_number("Longitude"),
    location_name = type_string("Name of the location for display to the user")
  )
)
}\if{html}{\out{</div>}}

Finally, we can extend \code{contents_shinychat()} to render our custom content
class for display in the chat interface. The basic process is to define a
\code{contents_shinychat()} external generic and then implement a method for your
custom class.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{contents_shinychat <- S7::new_external_generic(
  package = "shinychat",
  name = "contents_shinychat",
  dispatch_args = "contents"
)

S7::method(contents_shinychat, WeatherToolResult) <- function(content) \{
  # Your custom rendering logic here
\}
}\if{html}{\out{</div>}}

You can use this pattern to completely customize how the content is displayed
inside shinychat by returning HTML objects directly from this method.

You can also use this pattern to build upon the default shinychat display for
tool requests and results. By using \code{\link[S7:super]{S7::super()}}, you can create the
object shinychat uses for tool results (or tool requests), and then modify it
to suit your needs.

\if{html}{\out{<div class="sourceCode r">}}\preformatted{S7::method(contents_shinychat, WeatherToolResult) <- function(content) \{
  # Call the super method for ContentToolResult to get shinychat's defaults
  res <- contents_shinychat(S7::super(content, ContentToolResult))

  # Then update the result object with more specific content
  # In this case, we render the tool result dataframe as a \{gt\} table...
  res$value <- gt::as_raw_html(gt::gt(content@value))
  res$value_type <- "html"
  # ...and update the tool result title to include the location name
  res$title <- paste("Weather Forecast for", content@location_name)

  res
\}
}\if{html}{\out{</div>}}

Note that you do \strong{not} need to create a new class or extend
\code{contents_shinychat()} to customize the tool display. Rather, you can use the
strategies discussed in the \href{https://posit-dev.github.io/shinychat/r/articles/tool-ui.html}{Tool Calling UI article} to
customize the tool request and result display by providing a \code{display} list
in the \code{extra} argument of the tool result.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{chat_greeting}
\alias{chat_greeting}
\title{Create a greeting for a chat UI}
\usage{
chat_greeting(content, dismissible = TRUE)
}
\arguments{
\item{content}{The greeting content. Can be:
\itemize{
\item A string, interpreted as markdown.
\item An \code{\link[htmltools:HTML]{htmltools::HTML()}} object, rendered as raw HTML.
\item An htmltools tag or \code{\link[htmltools:tagList]{htmltools::tagList()}}, including Shiny
inputs/outputs.
\item A generator or promise (only valid when used with \code{chat_set_greeting()}).
}}

\item{dismissible}{Whether the greeting is automatically dismissed when the
user sends a message. Defaults to \code{TRUE}.}
}
\value{
An S3 object of class \code{"chat_greeting"}.
}
\description{
Creates a greeting object for use with \code{\link[=chat_ui]{chat_ui()}} or \code{\link[=chat_set_greeting]{chat_set_greeting()}}.
A greeting is displayed when the chat first loads and is dismissed when the
user sends their first message.
}
\section{Patterns}{


\strong{Non-dismissible greeting} (stays visible after the user sends a message):

\if{html}{\out{<div class="sourceCode r">}}\preformatted{chat_greeting("Please read our [terms of service](https://example.com).", dismissible = FALSE)
}\if{html}{\out{</div>}}

\strong{Greeting with suggestion cards} (clickable chips that fill the input):

\if{html}{\out{<div class="sourceCode r">}}\preformatted{chat_greeting(paste(
  "## Welcome!\\n\\n",
  "Try one of these:\\n\\n",
  '<span class="suggestion">Summarize my data</span>\\n',
  '<span class="suggestion">Create a plot</span>\\n',
  '<span class="suggestion">Explain this code</span>'
))
}\if{html}{\out{</div>}}

\strong{Greeting with HTML tags} (Shiny inputs/outputs):

\if{html}{\out{<div class="sourceCode r">}}\preformatted{chat_greeting(htmltools::tagList(
  htmltools::h2("Welcome!"),
  shiny::selectInput("model", "Choose a model:", c("gpt-4o", "claude-3"))
))
}\if{html}{\out{</div>}}
}

\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}
library(shiny)
library(bslib)
library(shinychat)

ui <- page_fillable(
  chat_ui(
    "chat",
    greeting = chat_greeting("## Welcome!\n\nHow can I help you today?")
  )
)

server <- function(input, output, session) {
  observeEvent(input$chat_user_input, {
    response <- paste0("You said: ", input$chat_user_input)
    chat_append("chat", response)
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chat.R
\name{chat_clear}
\alias{chat_clear}
\title{Clear all messages from a chat control}
\usage{
chat_clear(id, greeting = FALSE, session = getDefaultReactiveDomain())
}
\arguments{
\item{id}{The ID of the chat element}

\item{greeting}{If \code{TRUE}, also clears the greeting. When the greeting is
cleared, \code{greeting_requested} will fire again (if the chat is visible),
allowing the server to generate a new greeting.}

\item{session}{The Shiny session object}
}
\description{
Removes all messages from the chat UI. Set \code{greeting = TRUE} to also
clear the greeting, which re-triggers \code{greeting_requested} (see the
\strong{Greeting} section in \code{\link[=chat_ui]{chat_ui()}}).
}
\examples{
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}

library(shiny)
library(bslib)

ui <- page_fillable(
  chat_ui("chat", fill = TRUE),
  actionButton("clear", "Clear chat")
)

server <- function(input, output, session) {
  observeEvent(input$clear, {
    chat_clear("chat")
  })

  observeEvent(input$chat_user_input, {
    response <- paste0("You said: ", input$chat_user_input)
    chat_append("chat", response)
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
\dontshow{if (interactive()) withAutoprint(\{ # examplesIf}

library(shiny)
library(bslib)
library(shinychat)

# Regenerate greeting on clear
ui <- page_fillable(
  chat_ui("chat"),
  actionButton("new_chat", "New chat")
)

server <- function(input, output, session) {
  observeEvent(input$chat_greeting_requested, {
    chat_set_greeting("chat", "## Welcome!\n\nHow can I help?")
  })

  observeEvent(input$new_chat, {
    # Clearing with greeting = TRUE triggers greeting_requested again
    chat_clear("chat", greeting = TRUE)
  })

  observeEvent(input$chat_user_input, {
    chat_append("chat", paste0("You said: ", input$chat_user_input))
  })
}

shinyApp(ui, server)
\dontshow{\}) # examplesIf}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/shinychat-package.R
\docType{package}
\name{shinychat-package}
\alias{shinychat}
\alias{shinychat-package}
\title{shinychat: Chat UI Component for 'shiny'}
\description{
\if{html}{\figure{logo.svg}{options: style='float: right' alt='logo' width='120'}}

Provides a scrolling chat interface with multiline input, suitable for creating chatbot apps based on Large Language Models (LLMs). Designed to work particularly well with the 'ellmer' R package for calling LLMs.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://posit-dev.github.io/shinychat/r/}
  \item \url{https://github.com/posit-dev/shinychat}
  \item Report bugs at \url{https://github.com/posit-dev/shinychat/issues}
}

}
\author{
\strong{Maintainer}: Garrick Aden-Buie \email{garrick@adenbuie.com} (\href{https://orcid.org/0000-0002-7111-0077}{ORCID})

Authors:
\itemize{
  \item Garrick Aden-Buie \email{garrick@adenbuie.com} (\href{https://orcid.org/0000-0002-7111-0077}{ORCID})
  \item Joe Cheng \email{joe@posit.co}
  \item Carson Sievert \email{carson@posit.co}
  \item Barret Schloerke \email{barret@posit.co} (\href{https://orcid.org/0000-0001-9986-114X}{ORCID})
}

Other contributors:
\itemize{
  \item Posit Software, PBC (\href{https://ror.org/03wc8by49}{ROR}) [copyright holder, funder]
}

}
\keyword{internal}
